///////////////////////////////////////////////////////////////////////////////
//
// GameAppDoc.cpp
//
///////////////////////////////////////////////////////////////////////////////
#include "GameAppDoc.h"
#include "GameAppUi.h"
#include "GameResourceUtils.h"
#include "GameResources.h"
#include "GameUids.h"
#include "Game.pan"

#include <s32file.h>
#include <basched.h>
#include <barsread.h>
#include <coemain.h>
#include <coeutils.h>
#include <eikproc.h>
#include <eikapp.h>


///////////////////////////////////////////////////////////////////////////////
// CGameDocument
///////////////////////////////////////////////////////////////////////////////
CGameDocument::CGameDocument(CEikApplication& aApp, CCoeEnv& aEnv)
: CEikDocument(aApp), iEnv(aEnv)
	{
	}

///////////////////////////////////////////////////////////////////////////////
CGameDocument::~CGameDocument()
	{
	delete iModel;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameDocument::ConstructL()
	{
	const TSize KScreenSize(300, 100);
	const TDisplayMode KScreenMode(EColor4K);

	iModel = new(ELeave) CGameModel;
	iModel->ConstructL(iEnv.WsSession(), *(iEnv.ScreenDevice()), KScreenSize, KScreenMode);
	LoadHighScoresL();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameDocument::NewGameL()
	{
	LevelPathL(iFileName);
	TBuf<64> name;
	iEnv.ReadResource(name, R_ASTEROIDS_LEVEL_FILE_NAME);
	iFileName.Append(name);
	RFs& fs = Process()->FsSession();
	TUint a = 0U;
	TInt err = fs.Att(iFileName, a);
	if(err == KErrNone)
		{
		TRAP(err, OpenL(iFileName));
		if(err == KErrNoMemory)
			User::Leave(err);
		else if(err != KErrNone)
			(void)fs.Delete(iFileName);
		}
	if(err != KErrNone)
		{
		NewL(R_LEVEL_ASTEROIDS, iFileName);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameDocument::NewL(TInt aResourceId, const TDesC& aFileName)
	{
	// Load model from resource.
	iModel->LoadLevelL(iEnv, aResourceId);
	// Create, store, and switch to new file.
	RFs& fs = Process()->FsSession();
	(void)fs.MkDirAll(aFileName);
	CFileStore* store = CreateFileStoreLC(fs, aFileName);
	CleanupStack::Pop(store);
	ChangeFile(aFileName, store);
	UpdateState(EFalse);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameDocument::OpenL(const TDesC& aFileName)
	{
	CFileStore* store = iEikProcess->MainStore();
	const TBool reload = (aFileName == Process()->MainDocFileName());
	if(!reload)
		{
		store = CFileStore::OpenLC(Process()->FsSession(), aFileName, EFileShareExclusive|EFileWrite);
		}
	TStreamId rootId = store->Root();
	if((store->Type()[1] != KUidAppDllDoc) || (rootId == KNullStreamId))
		User::Leave(KErrCorrupt);
	CStreamDictionary* dictionary = CStreamDictionary::NewLC();
	RStoreReadStream root;
	root.OpenLC(*store, rootId);
	root >> *dictionary;
	CleanupStack::PopAndDestroy(); // root
	RestoreL(*store, *dictionary);
	CleanupStack::PopAndDestroy(dictionary);
	if(!reload)
		{
		CleanupStack::Pop(store);
		ChangeFile(aFileName, store);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameDocument::LevelPathL(TFileName& aFileName) const
	{
	HBufC* path = iEnv.AllocReadResourceLC(R_THRUST_T_LEVEL_PATH);
	Dll::FileName(aFileName);
	TParse p;
	User::LeaveIfError(p.SetNoWild(*path, &aFileName, NULL));
	aFileName = p.DriveAndPath();
	(void)Process()->FsSession().MkDirAll(aFileName);
	CleanupStack::PopAndDestroy(); // path
	}

///////////////////////////////////////////////////////////////////////////////
void CGameDocument::LoadHighScoresL()
	{
	CGameHighScoreTable* highScores = NULL;
	CDictionaryStore* iniFile = Application()->OpenIniFileLC(Process()->FsSession());
	if(iniFile->IsPresentL(KUidHighScoresIniFileStream))
		{
		TRAPD(err, highScores = ReadHighScoresStreamL(*iniFile));
		if(err != KErrNone)
			{
			TParse p;
			iEikApplication->SetToIniFileNameL(p);
			(void)Process()->FsSession().Delete(p.FullName());
			}
		}
	if(!highScores)
		{
		highScores = CGameHighScoreTable::NewL(KMaxNumScores);
		}
	CleanupStack::PopAndDestroy(iniFile);
	iModel->SetHighScores(highScores);
	}

///////////////////////////////////////////////////////////////////////////////
CGameHighScoreTable* CGameDocument::ReadHighScoresStreamL(CDictionaryStore& aStore)
	{
	RDictionaryReadStream stream;
	stream.OpenLC(aStore, KUidHighScoresIniFileStream);
	CGameHighScoreTable* highScores = CGameHighScoreTable::NewL(stream);
	CleanupStack::PopAndDestroy(); // stream
	return highScores;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameDocument::SaveHighScoresL()
	{
	CGameHighScoreTable* highScores = iModel->HighScores();
	if(highScores)
		{
		CDictionaryStore* iniFile = Application()->OpenIniFileLC(Process()->FsSession());
		RDictionaryWriteStream stream;
		stream.AssignLC(*iniFile, KUidHighScoresIniFileStream);
		highScores->ExternalizeL(stream);
		stream.CommitL();
		CleanupStack::PopAndDestroy(); // stream
		iniFile->CommitL();
		CleanupStack::PopAndDestroy(iniFile);
		}
	}

///////////////////////////////////////////////////////////////////////////////
CEikAppUi* CGameDocument::CreateAppUiL()
	{
	return new(ELeave) CGameAppUi;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameDocument::NewDocumentL()
	{
	// Reset model.
	__ASSERT_DEBUG(iModel, Panic(EPanicDocNotConstructed));
	iModel->ResetL();
	UpdateState(ETrue);
	}

//////////////////////////////////////////////////////////////////////////////
void CGameDocument::SaveL()
	{
	// Disable any framework saving with empty implementation.
	}

///////////////////////////////////////////////////////////////////////////////
void CGameDocument::StoreL(CStreamStore& aStore, CStreamDictionary& aStreamDic) const
	{
	// Store the model components to aStore.
	__ASSERT_DEBUG(iModel, Panic(EPanicDocNotConstructed));
	iModel->SaveLevelL(aStore, aStreamDic);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameDocument::RestoreL(const CStreamStore& aStore, const CStreamDictionary& aStreamDic)
	{
	// Restore the model components from aStore.
	__ASSERT_DEBUG(iModel, Panic(EPanicDocNotConstructed));
	iModel->LoadLevelL(aStore, aStreamDic);
	UpdateState(EFalse);
	}

//////////////////////////////////////////////////////////////////////////////
void CGameDocument::ChangeFile(const TDesC& aFileName, CFileStore* aStore)
	{
	// Change the apparc 'main store' for this app.
	CEikProcess* proc = iEikProcess;
	delete proc->MainStore();
	proc->SetMainStore(aStore);
	proc->SetMainDocFileName(aFileName);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameDocument::UpdateState(TBool aChanged)
	{
	// By this point we've updated the model, and if the app ui leaves
	// we can't roll-back anyway, so TRAP the error.
	TRAPD(ignore,iAppUi->HandleModelChangeL());
	SetChanged(aChanged);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
