///////////////////////////////////////////////////////////////////////////////
//
// ScreenColor.cpp
//
//
// 
//
//
///////////////////////////////////////////////////////////////////////////////
#include "ScreenColor.h"
#include "GameResourceUtils.h"
#include <barsread.h>


///////////////////////////////////////////////////////////////////////////////
// TFade
///////////////////////////////////////////////////////////////////////////////
TRgb TFade::FadeColor(TFade::TType aType, TRgb aFadeFrom, TRgb aFadeBy, TInt aSteps)
	{ // static
	TRgb c;
	switch(aType)
		{
	case ENoFade:
		c = aFadeFrom;
		break;
	case EDarker:
		c = TRgb( Max(0, aFadeFrom.Value() - (aFadeBy.Value() * aSteps)) );
		break;
	case EBrighter:
		c = TRgb( Min(255, aFadeFrom.Value() + (aFadeBy.Value() * aSteps)) );
		break;
		}
	return c;
	}

///////////////////////////////////////////////////////////////////////////////
TFade::TFade()
: iSteps(0U), iColStep(KRgbBlack), iFadeType((TUint8)ENoFade)
	{
	}

///////////////////////////////////////////////////////////////////////////////
TFade::TFade(TInt aSteps, TRgb aFadeFrom, TRgb aFadeTo)
: iSteps(0U), iColStep(KRgbBlack), iFadeType((TUint8)ENoFade)
	{
	Set(aSteps, aFadeFrom, aFadeTo);
	}

///////////////////////////////////////////////////////////////////////////////
void TFade::InternalizeL(TResourceReader& aReader)
	{
	const TInt steps = aReader.ReadUint8();
	const TRgb color1(aReader.ReadUint32());
	const TRgb color2(aReader.ReadUint32());
	Set(steps, color1, color2);
	}

///////////////////////////////////////////////////////////////////////////////
void TFade::InternalizeL(RReadStream& aStream)
	{
	iSteps = (TUint8)aStream.ReadUint8L();
	aStream >> iColStep;
	iFadeType = (TUint8)aStream.ReadUint8L();
	}

///////////////////////////////////////////////////////////////////////////////
void TFade::ExternalizeL(RWriteStream& aStream) const
	{
	aStream.WriteUint8L(iSteps);
	aStream << iColStep;
	aStream.WriteUint8L(iFadeType);
	}

///////////////////////////////////////////////////////////////////////////////
void TFade::Set(TInt aSteps, TRgb aStartCol, TRgb aEndCol)
	{
	iSteps = TUint8(aSteps);
	if(iSteps > 1)
		{
		// iColor2 is color increment.
		// Relative intensity determines fade direcition (brighter or darker).
		TRgb rgb;
		if(aStartCol.Difference(aEndCol) > 0)
			{
			iFadeType = TUint8(EDarker);
			rgb = TRgb(aStartCol.Value() - aEndCol.Value());
			}
		else
			{
			iFadeType = TUint8(EBrighter);
			rgb = TRgb(aEndCol.Value() - aStartCol.Value());
			}
		iColStep.SetRed(rgb.Red() / aSteps);
		iColStep.SetGreen(rgb.Green() / aSteps);
		iColStep.SetBlue(rgb.Blue() / aSteps);
		}
	else
		{
		iColStep = aEndCol;
		}
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// TFadedColor
///////////////////////////////////////////////////////////////////////////////
TFadedColor::TFadedColor()
: iFixedColor(KRgbWhite)
	{
	}

///////////////////////////////////////////////////////////////////////////////
TFadedColor::TFadedColor(TRgb aFixedColor)
: iFixedColor(aFixedColor)
	{
	Set(aFixedColor);
	}

///////////////////////////////////////////////////////////////////////////////
TFadedColor::TFadedColor(TInt aSteps, TRgb aStartCol, TRgb aEndCol)
: iFixedColor(KRgbWhite)
	{
	Set(aSteps, aStartCol, aEndCol);
	}

///////////////////////////////////////////////////////////////////////////////
void TFadedColor::InternalizeL(TResourceReader& aReader)
	{
	const TInt steps = aReader.ReadUint8();
	const TRgb color1(aReader.ReadUint32());
	const TRgb color2(aReader.ReadUint32());
	if(steps > 1)
		Set(steps, color1, color2);
	else
		Set(color1);
	}

///////////////////////////////////////////////////////////////////////////////
void TFadedColor::InternalizeL(RReadStream& aStream)
	{
	TFade::InternalizeL(aStream);
	aStream >> iFixedColor;
	}

///////////////////////////////////////////////////////////////////////////////
void TFadedColor::ExternalizeL(RWriteStream& aStream) const
	{
	TFade::ExternalizeL(aStream);
	aStream << iFixedColor;
	}

///////////////////////////////////////////////////////////////////////////////
void TFadedColor::Set(TRgb aFixedColor)
	{
	iSteps = 1;
	iFixedColor = aFixedColor;
	iFadeType = TUint8(ENoFade);
	}

///////////////////////////////////////////////////////////////////////////////
void TFadedColor::Set(TInt aSteps, TRgb aStartCol, TRgb aEndCol)
	{
	TFade::Set(aSteps, aStartCol, aEndCol);
	iFixedColor = aStartCol;
	}

///////////////////////////////////////////////////////////////////////////////
TRgb TFadedColor::Color(TInt aStep) const
	{
	return TFade::FadeColor(TFade::TType(iFadeType), iFixedColor, iColStep, aStep);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// TMultiFadedColor
///////////////////////////////////////////////////////////////////////////////
void TMultiFadedColor::InternalizeL(TResourceReader& aReader)
	{
	for(TInt c=0; c<KMaxFades; c++)
		{
		iFades[c].InternalizeL(aReader);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void TMultiFadedColor::InternalizeL(RReadStream& aStream)
	{
	for(TInt c=0; c<KMaxFades; c++)
		{
		iFades[c].InternalizeL(aStream);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void TMultiFadedColor::ExternalizeL(RWriteStream& aStream) const
	{
	for(TInt c=0; c<KMaxFades; c++)
		{
		iFades[c].ExternalizeL(aStream);
		}
	}

///////////////////////////////////////////////////////////////////////////////
TRgb TMultiFadedColor::Color(TInt aStep) const
	{
	TInt step = aStep;
	TInt c = 0;
	for(; c<KMaxFades; c++)
		{
		step -= iFades[c].Steps();
		if(step < 0)
			break;
		aStep -= iFades[c].Steps();
		}
	return iFades[c].Color(aStep);
	}

///////////////////////////////////////////////////////////////////////////////
TInt TMultiFadedColor::Steps() const
	{
	TInt steps = 0;
	for(TInt c=0; c<KMaxFades; c++)
		{
		steps += iFades[c].Steps();
		}
	return steps;
	}

///////////////////////////////////////////////////////////////////////////////
void TMultiFadedColor::Set(TInt aStepsForEachFade, const TRgb& aStartCol, const TRgb& aCol2, const TRgb& aCol3, const TRgb& aEndCol)
	{
	iFades[0].Set(aStepsForEachFade, aStartCol, aCol2);
	iFades[1].Set(aStepsForEachFade, aCol2, aCol3);
	iFades[2].Set(aStepsForEachFade, aCol3, aEndCol);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
