///////////////////////////////////////////////////////////////////////////////
//
// ScreenUtility.cpp
//
//
// 
//
//
///////////////////////////////////////////////////////////////////////////////
#include "ScreenUtility.h"
#include "GameResourceUtils.h"
#include <gulutil.h> // DrawUtils
#include <barsread.h>
#include <s32strm.h>

const TBool KInvertMaskSwitch = ETrue;
///////////////////////////////////////////////////////////////////////////////
// TBitmapDrawer
///////////////////////////////////////////////////////////////////////////////
void TBitmapDrawer::DrawBitmap(CBitmapContext& aGc, const CFbsBitmap& aBitmap, const TRect& aRect, const TRect* aSrcRect)
	{ // static
	TBitmapDrawer drawer(aBitmap);
	drawer.Draw(aGc, aRect, aSrcRect);
	}

///////////////////////////////////////////////////////////////////////////////
void TBitmapDrawer::BlitBitmap(CBitmapContext& aGc, const CFbsBitmap& aBitmap, const TPoint& aPos, const TRect* aSrcRect)
	{ // static
	const TRect srcRect(aBitmap.SizeInPixels());
	if(!aSrcRect)
		aSrcRect = &srcRect;
	TBitmapDrawer drawer(aBitmap);
	drawer.Blit(aGc, aPos, *aSrcRect);
	}

///////////////////////////////////////////////////////////////////////////////
TBitmapDrawer::TBitmapDrawer()
: iAlignment(EHCenterVCenter), iBitmap(NULL), iMaskBitmap(NULL)
	{
	}

///////////////////////////////////////////////////////////////////////////////
TBitmapDrawer::TBitmapDrawer(const CFbsBitmap& aBitmap, const CFbsBitmap* aMask)
: iAlignment(EHCenterVCenter), iBitmap(&aBitmap), iMaskBitmap(aMask)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void TBitmapDrawer::Set(const CFbsBitmap& aBitmap, const CFbsBitmap* aMask)
	{
	iBitmap = &aBitmap;
	iMaskBitmap = aMask;
	}

///////////////////////////////////////////////////////////////////////////////
void TBitmapDrawer::Reset()
	{
	iBitmap = NULL;
	iMaskBitmap = NULL;
	}

///////////////////////////////////////////////////////////////////////////////
void TBitmapDrawer::SetAlignment(TGulAlignmentValue aAlign)
	{
	iAlignment = aAlign;
	}

///////////////////////////////////////////////////////////////////////////////
void TBitmapDrawer::Draw(CBitmapContext& aGc, const TRect& aRect, const TRect* aSrcRect) const
	{
	if(!iBitmap)
		{
		aGc.DrawRect(aRect);
		return;
		}
	TRect srcRect;
	if(aSrcRect)
		srcRect = *aSrcRect;
	else
		srcRect = TRect(iBitmap->SizeInPixels());
	const TSize bmpSize(srcRect.Size());
	const TPoint bmpPos(iAlignment.InnerTopLeft(aRect, bmpSize));
	const TSize viewSize(aRect.Size());
	if( (viewSize.iWidth > bmpSize.iWidth) || (viewSize.iHeight > bmpSize.iHeight) )
		{
		aGc.SetPenStyle(CGraphicsContext::ENullPen);
		DrawUtils::DrawBetweenRects(aGc, aRect, TRect(bmpPos, bmpSize));
		}
	Blit(aGc, bmpPos, srcRect);
	}

///////////////////////////////////////////////////////////////////////////////
void TBitmapDrawer::Blit(CBitmapContext& aGc, const TPoint& aPos, const TRect& aSrcRect) const
	{
	ASSERT(iBitmap);
	if(iMaskBitmap)
		{
		aGc.BitBltMasked(aPos, iBitmap, aSrcRect, iMaskBitmap, KInvertMaskSwitch);
		}
	else
		{
		aGc.BitBlt(aPos, iBitmap, aSrcRect);
		}
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// TScreenObjectStyle
///////////////////////////////////////////////////////////////////////////////
void TScreenObjectStyle::UnSet(CGraphicsContext&)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void TScreenObjectStyle::InternalizeL(TResourceReader& aReader)
	{
	iStyle = (TInt)aReader.ReadUint8();
	iColor = TRgb(aReader.ReadUint32());
	iDrawMode = (TInt8)aReader.ReadInt8();
	}

///////////////////////////////////////////////////////////////////////////////
void TScreenObjectStyle::InternalizeL(RReadStream& aStream)
	{
	iStyle = (TInt)aStream.ReadUint8L();
	aStream >> iColor;
	iDrawMode = (TInt8)aStream.ReadInt8L();
	}

///////////////////////////////////////////////////////////////////////////////
void TScreenObjectStyle::ExternalizeL(RWriteStream& aStream) const
	{
	aStream.WriteUint8L(TUint(iStyle));
	aStream << iColor;
	aStream.WriteInt8L(iDrawMode);
	}

///////////////////////////////////////////////////////////////////////////////
TScreenObjectStyle::TScreenObjectStyle()
: iStyle(0),
	iColor(KRgbWhite),
	iDrawMode(-1)
	{
	}

///////////////////////////////////////////////////////////////////////////////
TScreenObjectStyle::TScreenObjectStyle(TInt aStyle, TRgb aRgb)
: iStyle(aStyle),
	iColor(aRgb),
	iDrawMode(-1)
	{
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// TScreenPenStyle
///////////////////////////////////////////////////////////////////////////////
TScreenPenStyle::TScreenPenStyle()
: TScreenObjectStyle(CGraphicsContext::ENullPen, KRgbBlack),
  iSize(1,1),
  iFont(NULL)
	{
	}

///////////////////////////////////////////////////////////////////////////////
TScreenPenStyle::TScreenPenStyle(CGraphicsContext::TPenStyle aStyle, TRgb aColor)
: TScreenObjectStyle(aStyle, aColor),
  iSize(1,1),
  iFont(NULL)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void TScreenPenStyle::Set(CGraphicsContext& aGc)
	{
	aGc.SetPenColor(iColor);
	aGc.SetPenStyle(CGraphicsContext::TPenStyle(iStyle));
	aGc.SetPenSize(iSize);
	if(iFont)
		{
		aGc.UseFont(iFont);
		}
	if(iDrawMode >= 0)
		{
		aGc.SetDrawMode(CGraphicsContext::TDrawMode(iDrawMode));
		}
	}

///////////////////////////////////////////////////////////////////////////////
void TScreenPenStyle::UnSet(CGraphicsContext& aGc)
	{
	if(iFont)
		{
		aGc.DiscardFont();
		}
	}

///////////////////////////////////////////////////////////////////////////////
void TScreenPenStyle::InternalizeL(TResourceReader& aReader)
	{
	TScreenObjectStyle::InternalizeL(aReader);
	iSize = GameResourceUtils::ReadTSize(aReader);
	}

///////////////////////////////////////////////////////////////////////////////
void TScreenPenStyle::InternalizeL(RReadStream& aStream)
	{
	TScreenObjectStyle::InternalizeL(aStream);
	aStream >> iSize;
	}

///////////////////////////////////////////////////////////////////////////////
void TScreenPenStyle::ExternalizeL(RWriteStream& aStream) const
	{
	TScreenObjectStyle::ExternalizeL(aStream);
	aStream << iSize;
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// TScreenFillStyle
///////////////////////////////////////////////////////////////////////////////
TScreenFillStyle::TScreenFillStyle()
: TScreenObjectStyle(CGraphicsContext::ENullBrush, KRgbWhite),
  iBitmap(NULL)
	{
	}

///////////////////////////////////////////////////////////////////////////////
TScreenFillStyle::TScreenFillStyle(CGraphicsContext::TBrushStyle aStyle, TRgb aColor)
: TScreenObjectStyle(aStyle, aColor),
  iBitmap(NULL)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void TScreenFillStyle::Set(CGraphicsContext& aGc)
	{
	aGc.SetBrushColor(iColor);
	aGc.SetBrushStyle(CGraphicsContext::TBrushStyle(iStyle));
	if(iBitmap)
		{
		aGc.UseBrushPattern(iBitmap);
		}
	if(iDrawMode >= 0)
		{
		aGc.SetDrawMode(CGraphicsContext::TDrawMode(iDrawMode));
		}
	}

///////////////////////////////////////////////////////////////////////////////
void TScreenFillStyle::UnSet(CGraphicsContext& aGc)
	{
	if(iBitmap)
		{
		aGc.DiscardBrushPattern();
		}
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
