///////////////////////////////////////////////////////////////////////////////
//
// GameModel.h
//
//
// The game model.
//
// This class provides the basic game operation, including providing the game
// tick, and loading and running levels.
//
///////////////////////////////////////////////////////////////////////////////
#ifndef GAMEMODEL_H
#define GAMEMODEL_H

#include <e32base.h>
#include <w32std.h>
#include <gdi.h>
#include "Sprites.h"
#include "Messages.h"
#include "GameModelObserver.h"

class CCoeEnv;
class CScreenBox;
class CScreenBitmap;
class CScreenBackground;
class CScreenMultilineTextBox;
class CGameObject;
class CGameObjectController;
class CGameHighScoreTable;

const TInt KGameMaxLevelNameLen = 64;

///////////////////////////////////////////////////////////////////////////////
// CGameModel
///////////////////////////////////////////////////////////////////////////////
class CGameModel : public CTimer
	{
public:
	CGameModel();
	virtual ~CGameModel();
	void ConstructL(RWsSession&, CWsScreenDevice&, const TSize& aScreenSize, TDisplayMode aScreenMode);
	void PlayLevelL(const CStreamStore&, const CStreamDictionary&);
	void LoadLevelL(const CCoeEnv& aCoeEnv, TInt aResourceId);
	void LoadLevelL(const CStreamStore&, const CStreamDictionary&);
	void SaveLevelL(CStreamStore&, CStreamDictionary&, TBool aCommit=ETrue) const;
	void ResetL();
	void SetName(const TDesC&);
	void SetHighScores(CGameHighScoreTable*);
public:
	void SetScreenSizeL(const TSize&);
	TInt AddScreenBoxL(CScreenBox*);
	void DeleteScreenBox(TInt);
	TInt ScreenBoxCount() const;
	CScreenBox& ScreenBox(TInt);
public:
	void SetGameObserver(MGameModelObserver& aOserver);
	void SetPlayerInput(MGameModelPlayerInput* aPlayerInput);
	void PlayL();
	void Stop();
	void ResumeL();
public:
	CGameHighScoreTable* HighScores();
	CScreenBitmap& Screen();
	CGameSpriteTable& Sprites();
	CGameObjectController& Objects();
	const TDesC& Name() const;
	TInt TicksPerSec() const;
	void SetTicksPerSec(TInt);
	void SendMessage(const TGameMessage&);

private:
	void CreateScreenL(const TSize&, TDisplayMode);
	void LoadDetailsL(TResourceReader&);
	void LoadDetailsL(RReadStream&);
	void LoadSpritesL(TResourceReader&);
	void LoadSpritesL(RReadStream&);
	void LoadObjectStatesL(TResourceReader&);
	void LoadObjectsL(TResourceReader&);
	void LoadObjectsL(RReadStream&);
	void LoadScreenL(TResourceReader&);
	void LoadScreenL(RReadStream&);
	void SaveDetailsL(RWriteStream&) const;
	void SaveSpritesL(RWriteStream&) const;
	void SaveObjectsL(RWriteStream&) const;
	void SaveScreenL(RWriteStream&) const;
private:
	virtual void RunL();
	void QueueGameTick();
	void NotifyObserverL(MGameModelObserver::TNotification);
private:
	void DrawBackground(CBitmapContext&);
	void DrawObjects(CBitmapContext&);
	void DrawScreenBoxes(CBitmapContext&);
private:
	RWsSession* iWs;
	CWsScreenDevice* iScreenDevice;
	CScreenBitmap* iScreen;
	CScreenBackground* iScreenBackground;
	RPointerArray<CScreenBox> iScreenBoxes;
private:
	TBuf<KGameMaxLevelNameLen> iName;
	TInt iGameTick;
	CGameSpriteTable* iSprites;
	CGameObjectController* iObjects;
	CGameHighScoreTable* iHighScores;
private:
	MGameModelObserver* iObserver;
	MGameModelPlayerInput* iPlayerInput;
	MGameModelPlayerInput::TPlayerKeyState iPlayerKeyState;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
#endif


