///////////////////////////////////////////////////////////////////////////////
//
// Messages.cpp
//
//
// 
//
//
///////////////////////////////////////////////////////////////////////////////
#include "Messages.h"
#include <s32strm.h>

const TGameMessage::TAddress TGameMessage::KNullAddress = 0;
const TGameMessage::TAddress TGameMessage::KBroadcastAddress = -1;

///////////////////////////////////////////////////////////////////////////////
// TGameMessage
///////////////////////////////////////////////////////////////////////////////
TGameMessage::TGameMessage()
: iId(KMsgId_Null),
  iSender(KNullAddress),
  iRecipient(KNullAddress),
  iState(0)
	{
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TGameMessage(TGameMessage::TId aId, TGameMessage::TAddress aSender, TGameMessage::TAddress aReceiver)
: iId(aId),
  iSender(aSender),
  iRecipient(aReceiver),
  iState(0)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void TGameMessage::InternalizeL(RReadStream& aStream)
	{
	iId = (TId)aStream.ReadUint16L();
	iSender = (TAddress)aStream.ReadInt32L();
	iRecipient = (TAddress)aStream.ReadInt32L();
	iState = aStream.ReadInt32L();
	aStream >> iParams;
	}

///////////////////////////////////////////////////////////////////////////////
void TGameMessage::ExternalizeL(RWriteStream& aStream) const
	{
	aStream.WriteUint16L(iId);
	aStream.WriteInt32L(iSender);
	aStream.WriteInt32L(iRecipient);
	aStream.WriteInt32L(iState);
	aStream << iParams;
	}

///////////////////////////////////////////////////////////////////////////////
void TGameMessage::LogText(TDes& aTxtBuf) const
	{
	_LIT(KFmt,"Message: %u From: %ld To: %ld StateId: %d Param Len: %ld");
	aTxtBuf.Format(KFmt, iId, iSender, iRecipient, iState, iParams.Length());
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// RGameMessageForwarder
///////////////////////////////////////////////////////////////////////////////
RGameMessageForwarder::RGameMessageForwarder()
	: iAddress(TGameMessage::KNullAddress),
		iHandler(NULL),
		iRouter(NULL)
	{
	}

///////////////////////////////////////////////////////////////////////////////
RGameMessageForwarder::~RGameMessageForwarder()
	{
	Close();
	}

///////////////////////////////////////////////////////////////////////////////
void RGameMessageForwarder::OpenL(MGameMessageHandler& aHandler, MGameMessageRouter& aRouter)
	{
	Close();
	aRouter.RegisterHandlerL(*this);
	iHandler = &aHandler;
	iRouter = &aRouter;
	}

///////////////////////////////////////////////////////////////////////////////
void RGameMessageForwarder::Close()
	{
	if(iHandler)
		{
		iRouter->UnRegisterHandler(*this);
		iAddress = TGameMessage::KNullAddress;
		iHandler = NULL;
		iRouter = NULL;
		}
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TAddress& RGameMessageForwarder::MessageAddress()
	{
	return iAddress;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse RGameMessageForwarder::OfferMessageL(const TGameMessage& aMessage)
	{
	return (iHandler) ? (iHandler->OfferMessageL(aMessage)) : (KGameMessageNotHandled);
	}

///////////////////////////////////////////////////////////////////////////////
void RGameMessageForwarder::RegisterHandlerL(MGameMessageHandler& aHandler)
	{
	if(iRouter)
		{
		iRouter->RegisterHandlerL(aHandler);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void RGameMessageForwarder::UnRegisterHandler(MGameMessageHandler& aHandler)
	{
	if(iRouter)
		{
		iRouter->UnRegisterHandler(aHandler);
		}
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse RGameMessageForwarder::SendMessage(const TGameMessage& aMessage)
	{
	return (iRouter) ? (iRouter->SendMessage(aMessage)) : (KGameMessageNotHandled);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// RGameMessageBroadcaster
///////////////////////////////////////////////////////////////////////////////
RGameMessageBroadcaster::RGameMessageBroadcaster()
: RGameMessageForwarder()
	{
	}

///////////////////////////////////////////////////////////////////////////////
RGameMessageBroadcaster::~RGameMessageBroadcaster()
	{
	Close();
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse RGameMessageBroadcaster::OfferMessageL(const TGameMessage& aMessage)
	{
	if(iHandler)
		{
		TGameMessage msg(aMessage);
		msg.iRecipient = TGameMessage::KBroadcastAddress;
		return iHandler->OfferMessageL(aMessage);
		}
	return KGameMessageNotHandled;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse RGameMessageBroadcaster::SendMessage(const TGameMessage& aMessage)
	{
	if(iRouter)
		{
		TGameMessage msg(aMessage);
		msg.iRecipient = TGameMessage::KBroadcastAddress;
		return iRouter->SendMessage(aMessage);
		}
	return KGameMessageNotHandled;
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CMessageCounterList
///////////////////////////////////////////////////////////////////////////////
CMessageCounterList::CMessageCounterList()
	{
	}

///////////////////////////////////////////////////////////////////////////////
CMessageCounterList::~CMessageCounterList()
	{
	iCounters.Reset();
	}

///////////////////////////////////////////////////////////////////////////////
TInt CMessageCounterList::AddCounter(const TMessageCounter& aCounter)
	{
	return iCounters.Append(aCounter);
	}

///////////////////////////////////////////////////////////////////////////////
void CMessageCounterList::MessageL(MGameMessageRouter& aRouter, const TGameMessage& aMessage)
	{
	TInt c = iCounters.Count();
	while(c--)
		{
		TMessageCounter& counter = iCounters[c];
		if(aMessage.iId == counter.iId)
			{
			--counter.iCount;
			if(counter.iCount == 0U)
				{
				(void)aRouter.SendMessage(counter.iExpiryMessage);
				iCounters.Remove(c);
				}
			}
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CMessageCounterList::InternalizeL(RReadStream& aStream)
	{
	iCounters.Reset();
	TInt c = (TInt)aStream.ReadUint16L();
	TMessageCounter counter;
	while(c--)
		{
		counter.iId = aStream.ReadUint16L();
		counter.iCount = aStream.ReadUint16L();
		counter.iExpiryMessage.InternalizeL(aStream);
		User::LeaveIfError(iCounters.Append(counter));
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CMessageCounterList::ExternalizeL(RWriteStream& aStream) const
	{
	const TInt num = iCounters.Count();
	aStream.WriteUint16L(TUint(num));
	for(TInt c=0; c<num; c++)
		{
		const TMessageCounter& counter = iCounters[c];
		aStream.WriteUint16L(counter.iId);
		aStream.WriteUint16L(counter.iCount);
		counter.iExpiryMessage.ExternalizeL(aStream);
		}
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
