///////////////////////////////////////////////////////////////////////////////
//
// MineState.cpp
//
//
///////////////////////////////////////////////////////////////////////////////
#include "MineState.h"
#include "ObjectController.h"
#include "GameResourceUtils.h"
#include <s32strm.h>
#include <barsread.h>

const TUid KUid0ObjectState = { KGameObjType0_ObjectState };

///////////////////////////////////////////////////////////////////////////////
// CGameObjectMineState
///////////////////////////////////////////////////////////////////////////////
CGameObjectMineState::CGameObjectMineState(CGameObjectController& aController)
: CGameObjectMotiveSolidState(aController)
	{
	}

///////////////////////////////////////////////////////////////////////////////
TUidType CGameObjectMineState::Type() const
	{
	const TUid uid = { KGameObjType1_ObjectMineState };
	return TUidType(KUid0ObjectState, uid, TUid::Uid(iId));
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectMineState::InternalizeL(TResourceReader& aReader)
	{
	CGameObjectMotiveSolidState::InternalizeL(aReader);
	iRange = GameResourceUtils::ReadTTrigUnit(aReader);
	iSpeed = GameResourceUtils::ReadTTrigUnit(aReader);
	iTargetAllegience = (TInt16)aReader.ReadInt16();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectMineState::InternalizeL(RReadStream& aStream)
	{
	CGameObjectMotiveSolidState::InternalizeL(aStream);
	iRange.InternalizeL(aStream);
	iSpeed.InternalizeL(aStream);
	iTargetAllegience = aStream.ReadInt16L();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectMineState::ExternalizeL(RWriteStream& aStream) const
	{
	CGameObjectMotiveSolidState::ExternalizeL(aStream);
	iRange.ExternalizeL(aStream);
	iSpeed.ExternalizeL(aStream);
	aStream.WriteInt16L(iTargetAllegience);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectMineState::Adjust(TCoOrds& aPosition, const TTrigXY& aDimensions, TMotionPhysics& aMotion)
	{
	// aPosition is TL.
	const TTrigUnit null;
	if(aPosition.iX < null)
		{
		AdjustXOutOfBounds(aPosition.iX, aDimensions.iX, ENegative, aMotion);
		}
	else if(aPosition.iX + aDimensions.iX > iExtents.iX)
		{
		AdjustXOutOfBounds(aPosition.iX, aDimensions.iX, EPositive, aMotion);
		}
	if(aPosition.iY < null)
		{
		AdjustYOutOfBounds(aPosition.iY, aDimensions.iY, ENegative, aMotion);
		}
	else if(aPosition.iY + aDimensions.iY > iExtents.iY)
		{
		AdjustYOutOfBounds(aPosition.iY, aDimensions.iY, EPositive, aMotion);
		}
	AdjustVelocity(aMotion);
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectMineState::OnEnterL(const TGameMessage& aMessage)
	{
	(void)CGameObjectMotiveSolidState::OnEnterL(aMessage);
	SetMaxSpeed(iController.ObjectMotionBounds().MaxSpeed());
	SetExtents(iController.ObjectMotionBounds().Extents());
	ObjectL(aMessage).SetMotionBounds(*this);
	return 1;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectMineState::OnUpdateL(const TGameMessage& aMessage)
	{
	(void)CGameObjectMotiveSolidState::OnUpdateL(aMessage);
	SeekTargetL(ObjectL(aMessage));
	return 1;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectMineState::AdjustXOutOfBounds(TTrigUnit& aPos, TTrigUnit aWidth, MMotionBounds::TOutOfBounds aOutOfBounds, TMotionPhysics& aMotion)
	{
	TVector v(aMotion.Velocity());
	TTrigUnit u(v.Magnitude());
	TDirection d;
	if(ENegative == aOutOfBounds)
		{
		aPos.Zero();
		// Set velocity to move vertically up screen.
		d = TDirection(0);
		}
	else
		{
		aPos = iExtents.iX - aWidth;
		// Set velocity to move vertically down screen.
		d = TDirection(TDirection::KResolution / 2);
		}
	v = d.Vector(u);
	aMotion.SetVelocity(v);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectMineState::AdjustYOutOfBounds(TTrigUnit& aPos, TTrigUnit aHeight, MMotionBounds::TOutOfBounds aOutOfBounds, TMotionPhysics& aMotion)
	{
	TVector v(aMotion.Velocity());
	TTrigUnit u(v.Magnitude());
	TDirection d(TDirection::KResolution / 4);
	if(ENegative == aOutOfBounds)
		{
		aPos.Zero();
		// Velocity moves right across the screen.
		}
	else
		{
		aPos = iExtents.iY - aHeight;
		// Set velocity to move left across the screen.
		d = TDirection(d.Value() * 3);
		}
	v = d.Vector(u);
	aMotion.SetVelocity(v);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectMineState::SeekTargetL(CGameObject& aMine)
	{
	TGameMessage::TAddress address;
	CGameObject* o = iController.FirstObject(address);
	while(o)
		{
		if( (o->Allegience() == iTargetAllegience) && (o != &aMine) )
			{
			TVector v(o->Centre());
			v -= aMine.Centre();
			if(v.Magnitude() <= iRange)
				{
				// Found a target.
				TDirection d = v.Direction();
				v = d.Vector(iSpeed);
				aMine.SetVelocity(v);
				break;
				}
			}
		o = iController.NextObject(address);
		}
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
