///////////////////////////////////////////////////////////////////////////////
//
// ObjectController.cpp
//
//
//
//
///////////////////////////////////////////////////////////////////////////////
#include "ObjectController.h"
#include "GameResourceUtils.h"
#include <e32math.h>

///////////////////////////////////////////////////////////////////////////////
//
// CGameObjectController
//
///////////////////////////////////////////////////////////////////////////////
CGameObjectController::CGameObjectController(CGameSpriteTable& aSprites)
: iSprites(aSprites),
  iObjectBounds(TTrigXY(100,100), 100),
  iAddress(TGameMessage::KNullAddress),
	iLastAddress(TGameMessage::KNullAddress)
	{
	ResetRandSeed();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectController::ConstructL(CGameObjectController::CObjectMap* aContainer, TBool aOwnsContainer,
																			 CGameObjectStateContainer* aStates, TBool aOwnsStates)
	{
	if( (iObjects) && (iFlags & EOwnsContainer) )
		{
		delete iObjects;
		iObjects = NULL;
		}
	if(aContainer)
		{
		iObjects = aContainer;
		if(aOwnsContainer)
			{
			iFlags |= EOwnsContainer;
			}
		else
			{
			iFlags &= ~EOwnsContainer;
			}
		}
	else
		{
		iObjects = new(ELeave) CObjectMap(*this);
		iFlags |= EOwnsContainer;
		iObjects->ResetL();
		}

	if( (iObjectStates) && (iFlags & EOwnsStates) )
		{
		delete iObjectStates;
		iObjectStates = NULL;
		}
	if(aStates)
		{
		iObjectStates = aStates;
		if(aOwnsStates)
			{
			iFlags |= EOwnsStates;
			}
		else
			{
			iFlags &= ~EOwnsStates;
			}
		}
	else
		{
		iObjectStates = new(ELeave) CGameObjectStateContainer(*this);
		iFlags |= EOwnsStates;
		iObjectStates->ResetL();
		}

	iAddress = 1;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectController::Reset()
	{
	TRAPD(err,
		ResetObjectsL(0);
		ResetStatesL(0);
		);
	ResetRandSeed();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectController::ResetObjectsL(TInt aSize)
	{
	iObjects->ResetL(aSize);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectController::ResetStatesL(TInt aSize)
	{
	iObjectStates->ResetL(aSize);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectController::ResetRandSeed()
	{
	TTime time;
	time.HomeTime();
	iRandSeed = time.Int64();
	iRandSeed *= iRandSeed;
	}

///////////////////////////////////////////////////////////////////////////////
CGameObjectController::~CGameObjectController()
	{
	Reset();
	if(iFlags & EOwnsContainer)
		{
		delete iObjects;
		}
	if(iFlags & EOwnsStates)
		{
		delete iObjectStates;
		}
	}

///////////////////////////////////////////////////////////////////////////////
// Message Handling.
///////////////////////////////////////////////////////////////////////////////
void CGameObjectController::RegisterHandlerL(MGameMessageHandler&)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectController::UnRegisterHandler(MGameMessageHandler& aHandler)
	{
	aHandler.MessageAddress() = TGameMessage::KNullAddress;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectController::SendMessage(const TGameMessage& aMessage)
	{
	TGameMessage::TResponse ret = KGameMessageNotHandled;
	TRAPD(err, ret = OfferMessageL(aMessage));
	return (err == KErrNone) ? (ret) : (err);
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TAddress& CGameObjectController::MessageAddress()
	{
	return iAddress;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectController::OfferMessageL(const TGameMessage& aMessage)
	{
	// Main message dispatch. All messages sent to game objects pass through here.
	// Note that this function may become re-entrant as OfferMessageL calls to game
	// objects may cause those objects to call SendMessage().
	TGameMessage::TResponse ret = KGameMessageNotHandled;
	if(IsBroadcastMessage(aMessage))
		{
		DispatchBroadcastMessageL(aMessage);
		ret = 1; // Handled.
		}
	else
		{
		// Find recipient.
		CGameObject* recipient = ObjectL(aMessage.iRecipient);
		if(recipient)
			{
			// Dispatch message.
			ret = DispatchMessageL(*recipient, aMessage);
			}
		// else, object has died, or address is invalid.
		}
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
// Object Add / Remove.
///////////////////////////////////////////////////////////////////////////////
void CGameObjectController::AddObjectL(CGameObject* aObject)
	{
	RegisterHandlerL(*aObject);
	TRAPD(err, iObjects->AddL(aObject));
	if(err != KErrNone)
		{
		UnRegisterHandler(*aObject);
		User::Leave(err);
		}
	// Enter initial state.
	const TGameMessage msg(KMsgId_ObjectEnterState, iAddress, aObject->MessageAddress());
	SendMessage(msg);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectController::DeleteObject(TGameMessage::TAddress aKey)
	{
	CGameObject* o = iObjects->Remove(aKey);
	if(o)
		{
		// Exit state and unregister the object.
		const TGameMessage msg(KMsgId_ObjectExitState, iAddress, aKey);
		SendMessage(msg);
		UnRegisterHandler(*o);
		// Destroy object.
		delete o;
		}
	}

///////////////////////////////////////////////////////////////////////////////
// Object Persistence.
///////////////////////////////////////////////////////////////////////////////
void CGameObjectController::InternalizeObjectStatesL(TResourceReader& aReader)
	{
	iObjectStates->InternalizeL(aReader);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectController::InternalizeObjectsL(TResourceReader& aReader)
	{
	iObjectBounds.SetMaxSpeed(GameResourceUtils::ReadTTrigUnit(aReader));
	iObjects->InternalizeL(aReader);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectController::InternalizeL(RReadStream& aStream)
	{
	iObjectStates->InternalizeL(aStream);
	TTrigUnit speed;
	speed.InternalizeL(aStream);
	iObjectBounds.SetMaxSpeed(speed);
	iObjects->InternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectController::ExternalizeL(RWriteStream& aStream) const
	{
	iObjectStates->ExternalizeL(aStream);
	const TTrigUnit speed(iObjectBounds.MaxSpeed());
	speed.ExternalizeL(aStream);
	iObjects->ExternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
// Object Draw.
///////////////////////////////////////////////////////////////////////////////
void CGameObjectController::Draw(CBitmapContext& aGc)
	{
	TInt32 key = 0;
	CGameObject* o = FirstObject(key);
	while(o)
		{
		o->Draw(aGc);
		o = NextObject(key);
		}
	}

///////////////////////////////////////////////////////////////////////////////
// Object Lookup Accessors.
///////////////////////////////////////////////////////////////////////////////
CGameObject* CGameObjectController::ObjectL(TGameMessage::TAddress aAddress) const
	{
	return iObjects->LookupL(aAddress);
	}

///////////////////////////////////////////////////////////////////////////////
CArrayPtrFlat<CGameObject>* CGameObjectController::ObjectsByStateLC(CGameObjectState::TId aState) const
	{
	CArrayPtrFlat<CGameObject>* objects = new(ELeave) CArrayPtrFlat<CGameObject>(10);
	CleanupStack::PushL(objects);
	TInt32 key = 0;
	CGameObject* o = FirstObject(key);
	while(o)
		{
		if(o->StateData().CurrentState() == aState)
			objects->AppendL(o);
		o = NextObject(key);
		}
	return objects;
	}

///////////////////////////////////////////////////////////////////////////////
CArrayPtrFlat<CGameObject>* CGameObjectController::ObjectsByAllegienceLC(TInt aAllegience, TIncludeExclude aIncExc) const
	{
	CArrayPtrFlat<CGameObject>* objects = new(ELeave) CArrayPtrFlat<CGameObject>(10);
	CleanupStack::PushL(objects);
	TInt32 key = 0;
	CGameObject* o = FirstObject(key);
	while(o)
		{
		if(aIncExc == EInclude)
			{
			if(o->Allegience() == aAllegience)
				objects->AppendL(o);
			}
		else
			{
			if(o->Allegience() != aAllegience)
				objects->AppendL(o);
			}
		o = NextObject(key);
		}
	return objects;
	}

///////////////////////////////////////////////////////////////////////////////
// Object Iteration Accessors.
///////////////////////////////////////////////////////////////////////////////
CGameObject* CGameObjectController::FirstObject(TGameMessage::TAddress& aAddress) const
	{
	return iObjects->First(aAddress);
	}

///////////////////////////////////////////////////////////////////////////////
CGameObject* CGameObjectController::NextObject(TGameMessage::TAddress& aAddress) const
	{
	return iObjects->Next(aAddress);
	}

///////////////////////////////////////////////////////////////////////////////
// State Accessors.
///////////////////////////////////////////////////////////////////////////////
CGameObjectState* CGameObjectController::StateL(CGameObjectState::TId aId) const
	{
	CGameObjectState* state = NULL;
	TInt c = iObjectStates->Count();
	while(c--)
		{
		state = &((*iObjectStates)[c]);
		if(state->Id() == aId)
			return state;
		}
	return NULL;
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
TInt CGameObjectController::Random(TInt aResolution)
	{
	TInt32 i = 0;
	if(aResolution > 1)
		{
		TReal r = Math::FRand(iRandSeed);
		r *= aResolution;
		(void)Math::Int(i, r);
		i = Min(aResolution - 1, i);
		}
	return i;
	}

///////////////////////////////////////////////////////////////////////////////
TInt CGameObjectController::Random(TInt aBase, TInt aRange, TInt aDelta)
	{
	TInt delta = Random(aRange + 1) - (aRange / 2);
	return aBase + (aDelta * delta);
	}

///////////////////////////////////////////////////////////////////////////////
TTrigUnit CGameObjectController::Random(const TTrigUnit& aBase, TInt aRange, const TTrigUnit& aDelta)
	{
	TInt delta = Random(aRange + 1) - (aRange / 2);
	return aBase + (aDelta * delta);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
void CGameObjectController::DispatchBroadcastMessageL(const TGameMessage& aMessage)
	{
	TGameMessage msg(aMessage);
	CGameObject* o = FirstObject(msg.iRecipient);
	while(o)
		{
		if(msg.iRecipient != msg.iSender)
			{
			(void)DispatchMessageL(*o, msg);
			}
		o = NextObject(msg.iRecipient);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectController::ProcessStateChangesAndDiscardDeadObjectsL()
	{
	TInt32 key = 0;
	CGameObject* o = FirstObject(key);
	while(o)
		{
		ProcessObjectStateChangeL(*o);
		if(o->StateData().CurrentState() == KStateIdDead)
			{
			DeleteObject(key);
			}
		o = NextObject(key);
		}
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectController::DispatchMessageL(CGameObject& aObject, const TGameMessage& aMessage)
	{
	TGameMessage::TResponse ret = aObject.OfferMessageL(aMessage);
//	(void)ProcessObjectStateChangeL(aObject);
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
TBool CGameObjectController::ProcessObjectStateChangeL(CGameObject& aObject)
	{
	// TODO: Leave safety.
	CGameObject::CStateData& stateData = aObject.StateData();
	if(stateData.StateChangeRequested())
		{
		const TGameMessage::TAddress address = aObject.MessageAddress();
		TGameMessage msg(KMsgId_ObjectExitState, iAddress, address);
		aObject.OfferMessageL(msg);
		stateData.ProcessStateChange();
		msg.iId = KMsgId_ObjectEnterState;
		aObject.OfferMessageL(msg);
		return ETrue;
		}
	return EFalse;
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
