///////////////////////////////////////////////////////////////////////////////
//
// ObjectController.h
//
//
//
//
///////////////////////////////////////////////////////////////////////////////
#ifndef OBJECTCONTROLLER_H
#define OBJECTCONTROLLER_H

#include "Objects.h"
#include "GameFactory.h"
#include "GameModelObserver.h" // TGameResult

//#define MOTION_BOUNDS_BOUNCE

///////////////////////////////////////////////////////////////////////////////
// CGameObjectController
///////////////////////////////////////////////////////////////////////////////
// Container for all CGameObject and derived class objects in the game.
// Delegates container management to a sub-object. This allows the container
// access stategy to be changed according to the expected volume of objects.
// Also contains all object states.
// Provides message routing, by being both a router and a receiver of messages.
// Message handler unique ids are assigned by this class to it's contained
// objects when they are added, and cleared when the objects are removed.
//
///////////////////////////////////////////////////////////////////////////////
class CGameObjectController : public CBase, public MGameMessageRouter, public MGameMessageHandler
	{
public:
	enum TIncludeExclude { EInclude, EExclude };
	class CObjectMap;
public:
	// Construct.
	CGameObjectController(CGameSpriteTable& aSprites);
	void ConstructL(CObjectMap* aContainer, TBool aOwnsContainer, CGameObjectStateContainer* aStates, TBool aOwnsStates);
	void Reset();
	void ResetObjectsL(TInt aSize);
	void ResetStatesL(TInt aSize);
	void ResetRandSeed();
	// Destruct.
	virtual ~CGameObjectController();
	// Persistence.
	void InternalizeObjectStatesL(TResourceReader&);
	void InternalizeObjectsL(TResourceReader&);
	void InternalizeL(RReadStream&);
	void ExternalizeL(RWriteStream&) const;
	// Message Handling.
	virtual TGameMessage::TResponse SendMessage(const TGameMessage&);
	virtual TGameMessage::TAddress& MessageAddress();
	virtual TGameMessage::TResponse OfferMessageL(const TGameMessage&);
	void ProcessStateChangesAndDiscardDeadObjectsL();
	// Object Add / Remove.
	void AddObjectL(CGameObject* aObject);
	void DeleteObject(TGameMessage::TAddress);
	// Object Draw.
	void Draw(CBitmapContext&);
	// Object Motion.
	inline void SetExtents(const TTrigXY&);
	inline void SetMaxSpeed(const TTrigUnit&);
	inline MMotionBounds& ObjectMotionBounds();
	// Object Lookup Accessors.
	CGameObject* ObjectL(TGameMessage::TAddress) const;
	CArrayPtrFlat<CGameObject>* ObjectsByStateLC(CGameObjectState::TId) const;
	CArrayPtrFlat<CGameObject>* ObjectsByAllegienceLC(TInt, TIncludeExclude aIncExc=EInclude) const;
	// Object Iteration Accessors.
	CGameObject* FirstObject(TGameMessage::TAddress&) const;
	CGameObject* NextObject(TGameMessage::TAddress&) const;
	// State Lookup Accessors.
	CGameObjectState* StateL(CGameObjectState::TId) const;
	inline CGameObjectStateContainer& ObjectStates();
	// Miscellaneous Accessors.
	inline CGameSpriteTable& Sprites();
	TInt Random(TInt aResolution=-1);
	TInt Random(TInt aBase, TInt aRange, TInt aDelta);
	TTrigUnit Random(const TTrigUnit& aBase, TInt aRange, const TTrigUnit& aDelta);
	void ClearResult();
protected:
	// Override the inherited access specification for these since this class manages
	// it's message handlers itself in AddObject() and DeleteObject().
	virtual void RegisterHandlerL(MGameMessageHandler&);
	virtual void UnRegisterHandler(MGameMessageHandler&);
private:
	inline TBool IsBroadcastMessage(const TGameMessage&) const;
	void DispatchBroadcastMessageL(const TGameMessage&);
	TGameMessage::TResponse DispatchMessageL(CGameObject&, const TGameMessage&);
	TBool ProcessObjectStateChangeL(CGameObject&);
private:
	enum TFlags                { EOwnsContainer=0x01, EOwnsStates=0x02 };
	TUint32                    iFlags;
	CGameSpriteTable&          iSprites; // Required mainly for InternalizeL
	CGameObjectStateContainer* iObjectStates;
	CObjectMap*                iObjects;
#ifdef MOTION_BOUNDS_BOUNCE
	TMotionBoundsBounce        iObjectBounds;
#else
	TMotionBoundsWrap          iObjectBounds;
#endif
	TInt64                     iRandSeed;
private:
	TGameMessage::TAddress     iAddress;
	TGameMessage::TAddress     iLastAddress;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CGameObjectController::CObjectMap
///////////////////////////////////////////////////////////////////////////////
class CGameObjectController::CObjectMap : public CGameContainerMap<CGameObject>
	{
public:
	CObjectMap(CGameObjectController& aController) : iController(aController) {}
protected:
	virtual MGameStreamable* NewByTypeLC(const TUidType& aType) { return GameFactory::NewObjectByTypeLC(aType, iController); }
protected:
	CGameObjectController& iController;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CGameObjectController inlines
///////////////////////////////////////////////////////////////////////////////
inline void CGameObjectController::SetExtents(const TTrigXY& aExtents)
	{ ObjectMotionBounds().SetExtents(aExtents); }
inline void CGameObjectController::SetMaxSpeed(const TTrigUnit& aSpeed)
	{ ObjectMotionBounds().SetMaxSpeed(aSpeed); }
inline MMotionBounds& CGameObjectController::ObjectMotionBounds()
	{ return iObjectBounds; }
inline CGameObjectStateContainer& CGameObjectController::ObjectStates()
	{ return *iObjectStates; }
inline CGameSpriteTable& CGameObjectController::Sprites()
	{ return iSprites; }
inline TBool CGameObjectController::IsBroadcastMessage(const TGameMessage& aMessage) const
	{ return ((aMessage.iRecipient == TGameMessage::KBroadcastAddress) || (aMessage.iRecipient == iAddress)); }

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
#endif
