///////////////////////////////////////////////////////////////////////////////
//
// ObjectStates.cpp
//
//
///////////////////////////////////////////////////////////////////////////////
#include "ObjectStates.h"
#include "ObjectController.h"
#include "GameTypes.hrh"
#include "GameResourceUtils.h"
#include <s32strm.h>
#include <barsread.h>

const TUid KUid0ObjectState = { KGameObjType0_ObjectState };

///////////////////////////////////////////////////////////////////////////////
// CGameObjectDeadState
///////////////////////////////////////////////////////////////////////////////
CGameObjectDeadState::CGameObjectDeadState(CGameObjectController& aController)
: CGameObjectState(aController, KGameObjType1_ObjectDeadState)
	{
	}

///////////////////////////////////////////////////////////////////////////////
TUidType CGameObjectDeadState::Type() const
	{
	const TUid uid = { KGameObjType1_ObjectDeadState };
	return TUidType(KUid0ObjectState, uid, TUid::Uid(iId));
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// CGameObjectActiveState
///////////////////////////////////////////////////////////////////////////////
CGameObjectActiveState::CGameObjectActiveState(CGameObjectController& aController)
: CGameObjectState(aController, KGameObjType1_ObjectActiveState),
  iLifetimeTotal(-1)
	{
	}

///////////////////////////////////////////////////////////////////////////////
TUidType CGameObjectActiveState::Type() const
	{
	const TUid uid = { KGameObjType1_ObjectActiveState };
	return TUidType(KUid0ObjectState, uid, TUid::Uid(iId));
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectActiveState::InternalizeL(TResourceReader& aReader)
	{
	CGameObjectState::InternalizeL(aReader);
	iFlags = aReader.ReadUint32();
	iLifetimeTotal = aReader.ReadInt16();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectActiveState::InternalizeL(RReadStream& aStream)
	{
	CGameObjectState::InternalizeL(aStream);
	iFlags = aStream.ReadUint32L();
	iLifetimeTotal = aStream.ReadInt16L();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectActiveState::ExternalizeL(RWriteStream& aStream) const
	{
	CGameObjectState::ExternalizeL(aStream);
	aStream.WriteUint32L(iFlags);
	aStream.WriteInt16L(iLifetimeTotal);
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectActiveState::OnEnterL(const TGameMessage& aMessage)
	{
	CGameObject& o = ObjectL(aMessage);
	if(iLifetimeTotal > 0)
		{
		AddTimedEventL(o, iLifetimeTotal, KMsgId_ObjectDie);
		}
	if(iFlags & EFadeSpriteWithLifetime)
		{
		AddTimedEventL(o, 1, KMsgId_ObjectIncSpriteFrame);
		}
	return ChangeObjectSpriteL(aMessage);
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectActiveState::OnExitL(const TGameMessage& aMessage)
	{
	return ChangeObjectSpriteL(aMessage);
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectActiveState::OnMessageL(const TGameMessage& aMessage)
	{
	TGameMessage::TResponse	ret = CGameObjectState::OnMessageL(aMessage);
	CGameObject& o = ObjectL(aMessage);
	switch(aMessage.iId)
		{
	case KMsgId_ObjectDie:
		Exit(o);
		break;
	case KMsgId_ObjectCollision:
		if(ObjectL(aMessage, CGameObjectState::ESender).Allegience() != KAllegienceNone)
			{
			ret = OnCollisionL(aMessage);
			}
		break;
	case KMsgId_ObjectIncSpriteFrame:
		o.IncSpriteFrame();
		AddTimedEventL(o, 1, KMsgId_ObjectIncSpriteFrame);
		break;
	default:
		break;
		}
	ChangeObjectSpriteL(aMessage);
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectActiveState::OnCollisionL(const TGameMessage& aMessage)
	{
	if(iFlags & EDieOnCollision)
		{
		Exit(ObjectL(aMessage));
		return 1;
		}
	return KGameMessageNotHandled;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectActiveState::ChangeObjectSpriteL(const TGameMessage&)
	{
	return KGameMessageNotHandled;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectActiveState::SendPointsScoreMessageL(const TGameMessage& aCollisionMessage)
	{
	CGameObject* o = &ObjectL(aCollisionMessage, CGameObjectState::ESender);
	TGameMessage msg(KMsgId_ObjectScorePoints, aCollisionMessage.iRecipient, TGameMessage::KBroadcastAddress);
	KMsgParamBuf_ScorePoints params;
	params().iAllegience = TInt16(o->Allegience());
	o = &ObjectL(aCollisionMessage);
	params().iPoints = o->Points();
	msg.iParams = params;
	iController.SendMessage(msg);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectActiveState::AddTimedEventL(CGameObject& aObject, TInt aCount, TGameMessage::TId aExpiryMsg)
	{
	const TGameMessage::TAddress address(aObject.MessageAddress());
	const TGameMessage msg(aExpiryMsg, address, address);
	AddMessageCounterL(aObject, KMsgId_ObjectUpdate, aCount, msg);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectActiveState::AddMessageCounterL(CGameObject& aObject, TGameMessage::TId aId, TInt aDuration, const TGameMessage& aMessage)
	{
	TMessageCounter msg;
	msg.iId = aId;
	msg.iCount = TUint16(aDuration);
	msg.iExpiryMessage = aMessage;
	User::LeaveIfError(aObject.StateData().AddCounter(msg));
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CGameObjectExplosionState
///////////////////////////////////////////////////////////////////////////////
CGameObjectExplosionState::CGameObjectExplosionState(CGameObjectController& aController)
: CGameObjectActiveState(aController),
  iStepLifetime(1)
	{
	iId = KGameObjType1_ObjectExplosionState;
	iLifetimeTotal = iStepLifetime * iStepColor.Steps();
	}

///////////////////////////////////////////////////////////////////////////////
TUidType CGameObjectExplosionState::Type() const
	{
	const TUid uid = { KGameObjType1_ObjectExplosionState };
	return TUidType(KUid0ObjectState, uid, TUid::Uid(iId));
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectExplosionState::InternalizeL(TResourceReader& aReader)
	{
	CGameObjectActiveState::InternalizeL(aReader);
	iSpriteId = aReader.ReadInt32();
	iStepLifetime = (TInt)aReader.ReadUint8();
	iStepColor.InternalizeL(aReader);
	iLifetimeTotal = iStepLifetime * iStepColor.Steps();
	iMinRadius = GameResourceUtils::ReadTTrigUnit(aReader);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectExplosionState::InternalizeL(RReadStream& aStream)
	{
	CGameObjectActiveState::InternalizeL(aStream);
	iSpriteId = aStream.ReadInt32L();
	iStepLifetime = (TInt)aStream.ReadUint8L();
	iStepColor.InternalizeL(aStream);
	iLifetimeTotal = iStepLifetime * iStepColor.Steps();
	iMinRadius.InternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectExplosionState::ExternalizeL(RWriteStream& aStream) const
	{
	CGameObjectActiveState::ExternalizeL(aStream);
	aStream.WriteInt32L(iSpriteId);
	aStream.WriteUint8L(TUint(iStepLifetime));
	iStepColor.ExternalizeL(aStream);
	iMinRadius.ExternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectExplosionState::OnEnterL(const TGameMessage& aMessage)
	{
	TGameMessage::TResponse ret = CGameObjectActiveState::OnEnterL(aMessage);
	CGameObject& o = ObjectL(aMessage);
	o.SetAllegience(KAllegienceNone);
	QueueAddRadialPointsMsgL(o);
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectExplosionState::OnMessageL(const TGameMessage& aMessage)
	{
	TGameMessage::TResponse	ret = CGameObjectActiveState::OnMessageL(aMessage);
	switch(aMessage.iId)
		{
	case KMsgId_ExplosionAddRadius:
		{
		CGameObject& o = ObjectL(aMessage);
		DeleteRadialPoints(o);
		AddRadialPointsL(o);
		QueueAddRadialPointsMsgL(o);
		}
		break;
	default:
		break;
		}
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectExplosionState::ChangeObjectSpriteL(const TGameMessage& aMessage)
	{
	if(aMessage.iId == KMsgId_ObjectEnterState)
		{
		CGameObject& o = ObjectL(aMessage);

		CGameSpriteData* pts = new(ELeave) CGameSpriteData(TDirection::KResolution);
		CleanupStack::PushL(pts);
		pts->SetColor(&iStepColor, EFalse, TDirection::KResolution);
		pts->iTrigUnit = o.Radius();
		pts->iTrigUnit /= iStepColor.Steps();
		if(pts->iTrigUnit < iMinRadius)
			pts->iTrigUnit = iMinRadius;
		pts->iTrigXY = o.Size();
		pts->iTrigXY /= 2;
		pts->AddRadialPointsL(pts->iTrigUnit, pts->iTrigXY);
		pts->iTrigUnit += pts->iTrigUnit;
		o.ChangeSpriteL(iSpriteId);
		o.SetSpriteData(pts, ETrue);
		CleanupStack::Pop(pts);

		o.MakeStationary();
		o.SetAllegience(KAllegienceNone);
		return 1;
		}
	return KGameMessageNotHandled;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectExplosionState::QueueAddRadialPointsMsgL(CGameObject& aObject)
	{
	AddTimedEventL(aObject, TUint16(iStepLifetime), KMsgId_ExplosionAddRadius);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectExplosionState::DeleteRadialPoints(CGameObject& aObject)
	{
	CGameSpriteData* spriteData = aObject.Sprite().Data();
	if( (spriteData) && (spriteData->Count() >= (TDirection::KResolution * (iStepColor.Steps() / 2))) )
		{
		TInt c = TDirection::KResolution;
		while(c--)
			{
			spriteData->DeletePoint(0);
			}
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectExplosionState::AddRadialPointsL(CGameObject& aObject)
	{
	CGameSpriteData* spriteData = aObject.Sprite().Data();
	if(spriteData)
		{
		spriteData->AddRadialPointsL(spriteData->iTrigUnit, spriteData->iTrigXY);
		spriteData->iTrigUnit += spriteData->iTrigUnit;
		}
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CGameObjectMotiveState
///////////////////////////////////////////////////////////////////////////////
CGameObjectMotiveState::CGameObjectMotiveState(CGameObjectController& aController)
: CGameObjectActiveState(aController)
	{
	iId = KGameObjType1_ObjectMotiveState;
	}

///////////////////////////////////////////////////////////////////////////////
TUidType CGameObjectMotiveState::Type() const
	{
	const TUid uid = { KGameObjType1_ObjectMotiveState };
	return TUidType(KUid0ObjectState, uid, TUid::Uid(iId));
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectMotiveState::OnUpdateL(const TGameMessage& aMessage)
	{
	(void)CGameObjectActiveState::OnUpdateL(aMessage);
	CGameObject& o = ObjectL(aMessage);
	o.UpdatePosition();
	o.UpdateFacing();
	return 1;
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CGameObjectMotiveState
///////////////////////////////////////////////////////////////////////////////
CGameObjectMotiveSolidState::CGameObjectMotiveSolidState(CGameObjectController& aController)
: CGameObjectMotiveState(aController)
	{
	iId = KGameObjType1_ObjectMotiveSolidState;
	}

///////////////////////////////////////////////////////////////////////////////
TUidType CGameObjectMotiveSolidState::Type() const
	{
	const TUid uid = { KGameObjType1_ObjectMotiveSolidState };
	return TUidType(KUid0ObjectState, uid, TUid::Uid(iId));
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectMotiveSolidState::OnUpdateL(const TGameMessage& aMessage)
	{
	(void)CGameObjectMotiveState::OnUpdateL(aMessage);
	return 1;
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
