///////////////////////////////////////////////////////////////////////////////
//
// ObjectStates.h
//
// Specific object states responsible for providing game logic by causing
// game objects to respond to particular messages.
//
// By default, object state IDs match their class type.
// N.B. that CGameObjectDeadState and CGameObjectPlayerState must always have 
// IDs that match their class type.
//
///////////////////////////////////////////////////////////////////////////////
#ifndef OBJECTSTATES_H
#define OBJECTSTATES_H

#include "State.h"
#include "Objects.h"
#include "ObjectEmitter.h"
#include "ScreenColor.h"


///////////////////////////////////////////////////////////////////////////////
// CGameObjectDeadState
///////////////////////////////////////////////////////////////////////////////
class CGameObjectDeadState : public CGameObjectState
	{
public:
	CGameObjectDeadState(CGameObjectController& aController);
	virtual TUidType Type() const;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// CGameObjectActiveState
///////////////////////////////////////////////////////////////////////////////
// Messages Handled:
//  KMsgId_ObjectUpdate
//  KMsgId_ObjectCollision
//
// Essentially a basic non-dead state, this class controls the object's
// lifetime by allowing a decay time, and optionally killing it on collision.
//
///////////////////////////////////////////////////////////////////////////////
class CGameObjectActiveState : public CGameObjectState
	{
public:
	CGameObjectActiveState(CGameObjectController& aController);
	virtual TUidType Type() const;
	virtual void InternalizeL(TResourceReader&);
	virtual void InternalizeL(RReadStream&);
	virtual void ExternalizeL(RWriteStream&) const;
protected:
	virtual TGameMessage::TResponse OnEnterL(const TGameMessage&);
	virtual TGameMessage::TResponse OnExitL(const TGameMessage&);
	virtual TGameMessage::TResponse OnMessageL(const TGameMessage&);
protected:
	virtual TGameMessage::TResponse OnCollisionL(const TGameMessage&);
	virtual TGameMessage::TResponse ChangeObjectSpriteL(const TGameMessage&);
protected:
	void SendPointsScoreMessageL(const TGameMessage& aCollisionMessage);
	void AddTimedEventL(CGameObject&, TInt aCount, TGameMessage::TId aExpiryMsg);
	void AddMessageCounterL(CGameObject&, TGameMessage::TId aCountMsg, TInt aCount, const TGameMessage& aExpiryMsg);
protected:
	enum TFlags { EDieOnCollision = 0x01, EFadeSpriteWithLifetime = 0x02, ENoUpdateLifetime = 0x04 };
	TUint32 iFlags;
	TInt iLifetimeTotal;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CGameObjectExplosionState
///////////////////////////////////////////////////////////////////////////////
class CGameObjectExplosionState : public CGameObjectActiveState
	{
public:
	CGameObjectExplosionState(CGameObjectController& aController);
	virtual TUidType Type() const;
	virtual void InternalizeL(TResourceReader&);
	virtual void InternalizeL(RReadStream&);
	virtual void ExternalizeL(RWriteStream&) const;
protected:
	virtual TGameMessage::TResponse OnEnterL(const TGameMessage&);
	virtual TGameMessage::TResponse OnMessageL(const TGameMessage&);
	virtual TGameMessage::TResponse ChangeObjectSpriteL(const TGameMessage&);
protected:
	void QueueAddRadialPointsMsgL(CGameObject&);
	void DeleteRadialPoints(CGameObject&);
	void AddRadialPointsL(CGameObject&);
protected:
	TInt iSpriteId;
	TInt iStepLifetime;
	TFadedColor iStepColor;
	TTrigUnit iMinRadius;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CGameObjectMotiveState
///////////////////////////////////////////////////////////////////////////////
// Messages Handled:
//  KMsgId_ObjectUpdate
//  KMsgId_ObjectCollision
//
// Updates object position and facing based on motion.
//
///////////////////////////////////////////////////////////////////////////////
class CGameObjectMotiveState : public CGameObjectActiveState
	{
public:
	CGameObjectMotiveState(CGameObjectController& aController);
	virtual TUidType Type() const;

protected:
	virtual TGameMessage::TResponse OnUpdateL(const TGameMessage&);
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CGameObjectMotiveSolidState
///////////////////////////////////////////////////////////////////////////////
// Messages Handled:
//  KMsgId_ObjectUpdate
//  KMsgId_ObjectCollision
//
// Updates object position and facing based on motion.
// Checks for collisions.
//
///////////////////////////////////////////////////////////////////////////////
class CGameObjectMotiveSolidState : public CGameObjectMotiveState
	{
public:
	CGameObjectMotiveSolidState(CGameObjectController& aController);
	virtual TUidType Type() const;

protected:
	virtual TGameMessage::TResponse OnUpdateL(const TGameMessage&);
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
#endif
