///////////////////////////////////////////////////////////////////////////////
//
// SpriteBitmap.cpp
//
//
///////////////////////////////////////////////////////////////////////////////
#include "SpriteBitmap.h"
#include "ScreenUtility.h"
#include "Motion.h"
#include "Game.pan"
#include "GameTypes.hrh"
#include <s32strm.h>
#include <barsread.h>

const TUid KUidSprite       = { KGameObjType0_Sprite };
const TUid KUidBitmapSprite = { KGameObjType1_SpriteBitmap };

///////////////////////////////////////////////////////////////////////////////
// CBitmapSpriteFrame
///////////////////////////////////////////////////////////////////////////////
CBitmapSpriteFrame::CBitmapSpriteFrame(RWsSession& aWs)
: iWs(aWs)
	{
	}

///////////////////////////////////////////////////////////////////////////////
CBitmapSpriteFrame::~CBitmapSpriteFrame()
	{
	if(iBitmap)
		{
		iBitmap->Reset();
		delete iBitmap;
		}
	if(iMask)
		{
		iMask->Reset();
		delete iMask;
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CBitmapSpriteFrame::InternalizeL(RReadStream& aStream)
	{
	ResetL();
	iBitmap->InternalizeL(aStream);
	iMask->InternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CBitmapSpriteFrame::ExternalizeL(RWriteStream& aStream) const
	{
	__ASSERT_DEBUG( (iBitmap) && (iMask), Panic(EPanicSpriteBitmapNoBitmaps));
	iBitmap->ExternalizeL(aStream);
	iMask->ExternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CBitmapSpriteFrame::ResetL()
	{
	if(iBitmap)
		{
		iBitmap->Reset();
		}
	else
		{
		iBitmap = new(ELeave) CBitmap(iWs);
		}
	if(iMask)
		{
		iMask->Reset();
		}
	else
		{
		iMask = new(ELeave) CBitmap(iWs);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CBitmapSpriteFrame::LoadL(CWsScreenDevice& aScreenDevice, const TDesC& aFile, TInt aFrame, TInt aMask)
	{
	LoadAndMapBitmapL(iBitmap, aScreenDevice, aFile, aFrame);
	LoadAndMapBitmapL(iMask, aScreenDevice, aFile, aMask);
	}

///////////////////////////////////////////////////////////////////////////////
void CBitmapSpriteFrame::Drawer(TBitmapDrawer& aDrawer) const
	{
	__ASSERT_DEBUG( (iBitmap) && (iMask), Panic(EPanicSpriteBitmapNoBitmaps));
	aDrawer.Set(*iBitmap, iMask);
	}

///////////////////////////////////////////////////////////////////////////////
TSize CBitmapSpriteFrame::Size() const
	{
	TSize size;
	if(iBitmap)
		{
		size = iBitmap->SizeInPixels();
		}
	return size;
	}

///////////////////////////////////////////////////////////////////////////////
void CBitmapSpriteFrame::LoadAndMapBitmapL(CBitmap*& aTarget, CWsScreenDevice& aScreenDevice, const TDesC& aFile, TInt aFileIndex)
	{
	CFbsBitmap* bitmap = new(ELeave) CFbsBitmap;
	CleanupStack::PushL(bitmap);
	User::LeaveIfError( bitmap->Load(aFile, aFileIndex) );
	const TSize sizeInPixels(bitmap->SizeInPixels());

	CScreenBitmap* mapper = new(ELeave) CScreenBitmap;
	CleanupStack::PushL(mapper);
	mapper->ConstructL(iWs, bitmap->SizeInPixels(), aScreenDevice.DisplayMode());
	TBitmapDrawer::DrawBitmap(*(mapper->Context()), *bitmap, TRect(sizeInPixels));

	CBitmap* mappedBitmap = mapper->ResetReleaseBitmapLC(); // Will fail to compile if bitmap types are not compatible.
	delete aTarget;
	aTarget = mappedBitmap;
	
	CleanupStack::Pop(mappedBitmap);
	CleanupStack::PopAndDestroy(mapper);
	CleanupStack::PopAndDestroy(bitmap);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// CGameBitmapSprite
///////////////////////////////////////////////////////////////////////////////
CGameBitmapSprite::CGameBitmapSprite(RWsSession& aWs, CWsScreenDevice& aScreenDevice)
: iWs(aWs), iScreenDevice(aScreenDevice)
	{
	}

///////////////////////////////////////////////////////////////////////////////
CGameBitmapSprite::~CGameBitmapSprite()
	{
	if(iFrames)
		{
		iFrames->ResetAndDestroy();
		delete iFrames;
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameBitmapSprite::ResetL()
	{
	if(iFrames)
		{
		iFrames->ResetAndDestroy();
		}
	else
		{
		iFrames = new(ELeave) CArrayPtrFlat<CBitmapSpriteFrame>(4);
		}
	}

///////////////////////////////////////////////////////////////////////////////
TUidType CGameBitmapSprite::Type() const
	{
	return TUidType(KUidSprite, KUidBitmapSprite);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameBitmapSprite::SetId(TInt32)
	{
	}

///////////////////////////////////////////////////////////////////////////////
TInt CGameBitmapSprite::FrameCount(CGameSpriteData*) const
	{
	__ASSERT_DEBUG(iFrames, Panic(EPanicSpriteBitmapNoFrames));
	return iFrames->Count();
	}

///////////////////////////////////////////////////////////////////////////////
TSize CGameBitmapSprite::FrameSize(TInt aFrame, CGameSpriteData*) const
	{
	__ASSERT_DEBUG(iFrames, Panic(EPanicSpriteBitmapNoFrames));
	return iFrames->At(aFrame)->Size();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameBitmapSprite::DrawFrame(TInt aFrame, CBitmapContext& aGc, const TPoint& aPos, CGameSpriteData*, MMotionBounds* aBounds) const
	{
	__ASSERT_DEBUG(iFrames, Panic(EPanicSpriteBitmapNoFrames));
	CBitmapSpriteFrame* frame = iFrames->At(aFrame);
	TBitmapDrawer drawer;
	frame->Drawer(drawer);
	aGc.Reset();
	const TSize frameSize(frame->Size());
	TRect srcRect(frameSize);
	TRect trgRect(aPos, frameSize);
	TBool drawn = EFalse;
	if(aBounds)
		{
		const TSize extents(aBounds->Extents());
		if(trgRect.iTl.iX < 0)
			{
			srcRect.iTl.iX -= trgRect.iTl.iX;
			trgRect.iTl.iX = 0;
			drawer.Draw(aGc, trgRect, &srcRect);
			srcRect.iBr.iX = srcRect.iTl.iX;
			srcRect.iTl.iX = 0;
			trgRect.iTl.iX = extents.iWidth - srcRect.Width();
			trgRect.iBr.iX = extents.iWidth;
			drawer.Draw(aGc, trgRect, &srcRect);
			drawn = ETrue;
			}
		else if(trgRect.iBr.iX > extents.iWidth)
			{
			srcRect.iBr.iX -= (trgRect.iBr.iX - extents.iWidth);
			trgRect.iBr.iX = extents.iWidth;
			trgRect.iTl.iX = trgRect.iBr.iX - srcRect.Width();
			drawer.Draw(aGc, trgRect, &srcRect);
			srcRect.iTl.iX = srcRect.iBr.iX;
			srcRect.iBr.iX = frameSize.iWidth;
			trgRect.iTl.iX = 0;
			trgRect.iBr.iX = srcRect.Width();
			drawer.Draw(aGc, trgRect, &srcRect);
			drawn = ETrue;
			}
		if(trgRect.iTl.iY < 0)
			{
			srcRect.iTl.iY -= trgRect.iTl.iY;
			trgRect.iTl.iY = 0;
			drawer.Draw(aGc, trgRect, &srcRect);
			srcRect.iBr.iY = srcRect.iTl.iY;
			srcRect.iTl.iY = 0;
			trgRect.iTl.iY = extents.iHeight - srcRect.Height();
			trgRect.iBr.iY = extents.iHeight;
			drawer.Draw(aGc, trgRect, &srcRect);
			drawn = ETrue;
			}
		else if(trgRect.iBr.iY > extents.iHeight)
			{
			srcRect.iBr.iY -= (trgRect.iBr.iY - extents.iHeight);
			trgRect.iBr.iY = extents.iHeight;
			trgRect.iTl.iY = trgRect.iBr.iY - srcRect.Height();
			drawer.Draw(aGc, trgRect, &srcRect);
			srcRect.iTl.iY = srcRect.iBr.iY;
			srcRect.iBr.iY = frameSize.iHeight;
			trgRect.iTl.iY = 0;
			trgRect.iBr.iY = srcRect.Height();
			drawer.Draw(aGc, trgRect, &srcRect);
			drawn = ETrue;
			}
		}
	if(!drawn)
		{
		drawer.Draw(aGc, trgRect, &srcRect);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameBitmapSprite::InternalizeL(TResourceReader& aReader)
	{
	ResetL();
	TInt numFrames = (TInt)aReader.ReadUint8();
	const TInt numMasks = (TInt)aReader.ReadUint8();
	__ASSERT_ALWAYS((numMasks == numFrames) || (numMasks == 1), Panic(EPanicSpriteBitmapBadMaskCount));
	TParse p;
	ResolveMbmNameL(p, aReader.ReadTPtrC()); // Mbm file name.
	const TPtrC mbmName(p.FullName());
	TInt idx = aReader.ReadUint8();     // Mbm index of first frame.
	TInt maskIdx = aReader.ReadUint8(); // Mbm index of first mask.
	numFrames += idx;                   // Find the index beyond the last frame.
	for(; idx<numFrames; idx++)
		{
		CBitmapSpriteFrame* frame = new(ELeave) CBitmapSpriteFrame(iWs);
		CleanupStack::PushL(frame);
		frame->LoadL(iScreenDevice, mbmName, idx, maskIdx);
		AppendFrameL(frame);
		CleanupStack::Pop(); // frame
		if(numMasks != 1)
			++maskIdx;
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameBitmapSprite::ResolveMbmNameL(TParse& aParse, const TDesC& aMbmNameAndPath)
	{
	TFileName dllName;
	Dll::FileName(dllName);
	User::LeaveIfError(aParse.Set(aMbmNameAndPath, &dllName, NULL));
	}

///////////////////////////////////////////////////////////////////////////////
void CGameBitmapSprite::InternalizeL(RReadStream& aStream)
	{
	ResetL();
	TInt count = aStream.ReadInt16L();
	while(count--)
		{
		CBitmapSpriteFrame* frame = new(ELeave) CBitmapSpriteFrame(iWs);
		CleanupStack::PushL(frame);
		frame->InternalizeL(aStream);
		AppendFrameL(frame);
		CleanupStack::Pop();
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameBitmapSprite::ExternalizeL(RWriteStream& aStream) const
	{
	__ASSERT_DEBUG(iFrames, Panic(EPanicSpriteBitmapNoFrames));
	const TInt count = FrameCount(NULL);
	aStream.WriteInt16L(count);
	for(TInt id=0; id<count; id++)
		{
		iFrames->At(id)->ExternalizeL(aStream);
		}
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
