///////////////////////////////////////////////////////////////////////////////
//
// Sprites.cpp
//
//
///////////////////////////////////////////////////////////////////////////////
#include "Sprites.h"
#include "SpriteTable.h"
#include "ScreenColor.h"
#include "GameFactory.h"
#include "GameResourceUtils.h"
#include "Motion.h"
#include "Game.pan"
#include <gdi.h>
#include <s32strm.h>
#include <barsread.h>

	
///////////////////////////////////////////////////////////////////////////////
// CGameSpriteData
///////////////////////////////////////////////////////////////////////////////
CGameSpriteData::CGameSpriteData()
: iColorGradient(1)
	{
	}

///////////////////////////////////////////////////////////////////////////////
CGameSpriteData::CGameSpriteData(TInt aVertexArrayGranularity)
: iPoints(aVertexArrayGranularity), iColorGradient(1)
	{
	}

///////////////////////////////////////////////////////////////////////////////
CGameSpriteData::~CGameSpriteData()
	{
	ResetPoints();
	if(iFlags & EFlagOwnsColor)
		{
		delete iColor;
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameSpriteData::InternalizeL(TResourceReader& aReader)
	{
	ResetPoints();
	TInt c = aReader.ReadInt16();
	while(c--)
		{
		User::LeaveIfError(iPoints.Append(GameResourceUtils::ReadTPoint(aReader)));
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameSpriteData::InternalizeL(RReadStream& aStream)
	{
	ResetPoints();
	TPoint p;
	TInt x = (TInt)aStream.ReadUint8L();
	while(x--)
		{
		aStream >> p;
		User::LeaveIfError(iPoints.Append(p));
		}
	iTrigUnit.InternalizeL(aStream);
	iTrigXY.InternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameSpriteData::ExternalizeL(RWriteStream& aStream) const
	{
	TInt x = Count();
	aStream.WriteUint8L(TUint(x));
	TInt y = 0;
	for(; y<x; y++)
		{
		aStream << iPoints[y];
		}
	iTrigUnit.ExternalizeL(aStream);
	iTrigXY.ExternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameSpriteData::PlotPoints(CGraphicsContext& aGc, const TPoint& aStart, MMotionBounds*)
	{
	if(iColorGradient < 1)
		iColorGradient = 1;
	aGc.SetOrigin();
	TPoint p;
	const TInt numPoints = Count();
	for(TInt c=0; c<numPoints; c++)
		{
		if(iColor)
			{
			aGc.SetPenColor(iColor->Color(c / iColorGradient));
			}
		p = aStart + Point(c);
		aGc.Plot(p);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameSpriteData::PlotPolygon(CGraphicsContext& aGc, const TPoint& aStart, MMotionBounds* aBounds)
	{
	const TInt numPoints = Count();
	if(!numPoints)
		return;

	const TPoint first(ProcessVertex(aStart, Point(0), aBounds));
	aGc.MoveTo(first);
	TPoint p;
	for(TInt c=1; c<numPoints; c++)
		{
		if(iColor)
			aGc.SetPenColor(iColor->Color(c));
		p = ProcessVertex(aStart, Point(c), aBounds);
		aGc.DrawLineTo(p);
		}
	aGc.DrawLineTo(first);
	}

///////////////////////////////////////////////////////////////////////////////
TPoint CGameSpriteData::ProcessVertex(const TPoint& aStart, const TPoint& aPos, MMotionBounds* aBounds)
	{
	TCoOrds p;
	p += aStart;
	p += aPos;
	if(aBounds)
		{
		TMotionPhysics motion;
		TMotionBoundsClip clipper(aBounds->Extents(), TTrigUnit(0));
		clipper.Adjust(p, TTrigXY(), motion);
		}
	return p;
	}

///////////////////////////////////////////////////////////////////////////////
TInt CGameSpriteData::Count() const
	{
	return iPoints.Count();
	}

///////////////////////////////////////////////////////////////////////////////
const TPoint& CGameSpriteData::Point(TInt aIndex) const
	{
	return iPoints[aIndex];
	}

///////////////////////////////////////////////////////////////////////////////
void CGameSpriteData::SetPointL(TInt aIndex, const TPoint& aPoint)
	{
	while(aIndex >= iPoints.Count())
		{
		User::LeaveIfError( iPoints.Append(TPoint()) );
		}
	iPoints[aIndex] = aPoint;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameSpriteData::InsertPointL(TInt aIndex, const TPoint& aPoint)
	{
	User::LeaveIfError(iPoints.Insert(aPoint, aIndex));
	}

///////////////////////////////////////////////////////////////////////////////
void CGameSpriteData::DeletePoint(TInt aIndex)
	{
	iPoints.Remove(aIndex);
	}

///////////////////////////////////////////////////////////////////////////////
TInt CGameSpriteData::AppendPointL(const TPoint& aPoint)
	{
	const TInt ret = Count();
	User::LeaveIfError(iPoints.Append(aPoint));
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameSpriteData::ResetPoints()
	{
	iPoints.Reset();
	}

///////////////////////////////////////////////////////////////////////////////
TRect CGameSpriteData::Bounds() const
	{
	TRect bounds;
	TInt c = Count();
	while(c--)
		{
		const TPoint& p = iPoints[c];
		bounds.iTl.iX = Min(bounds.iTl.iX, p.iX);
		bounds.iBr.iX = Max(bounds.iBr.iX, p.iX);
		bounds.iTl.iY = Min(bounds.iTl.iY, p.iY);
		bounds.iBr.iY = Max(bounds.iBr.iY, p.iY);
		}
	return bounds;
	}

///////////////////////////////////////////////////////////////////////////////
MDynamicColor* CGameSpriteData::Color()
	{
	return iColor;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameSpriteData::SetColor(MDynamicColor* aColor, TBool aOwned, TInt aGradient)
	{
	if(iFlags & EFlagOwnsColor)
		{
		delete iColor;
		iFlags &= ~EFlagOwnsColor;
		}
	iColor = aColor;
	if(aOwned)
		{
		iFlags |= EFlagOwnsColor;
		}
	iColorGradient = aGradient;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameSpriteData::AddRadialPointsL(TTrigUnit aRadius, const TPoint& aCentreOffset)
	{
	TDirection a;
	TPoint p;
	AppendPointL(aCentreOffset);
	do
		{
		p = a.Vector(aRadius);
		AppendPointL(aCentreOffset + p);
		++a;
		}while(a.Value() != 0);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameSpriteData::AddTrianglePointsL(TTrigUnit aRadius)
	{
	const TInt r = aRadius.NearestInt();
	const TInt rby2 = TTrigUnit(aRadius / 2).NearestInt();
	TPoint p(rby2, 0);
	AppendPointL(p);
	p.iX = rby2;
	p.iY = r;
	AppendPointL(p);
	p.iX = -r;
	p.iY = 0;
	AppendPointL(p);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// RGameSprite
///////////////////////////////////////////////////////////////////////////////
RGameSprite::RGameSprite()
: iId(-1), iSpriteTable(NULL), iData(NULL), iOwnsData(EFalse), iSpriteFrame(0)
	{
	}

///////////////////////////////////////////////////////////////////////////////
RGameSprite::RGameSprite(CGameSpriteTable& aSpriteTable)
: iId(-1), iSpriteTable(&aSpriteTable), iData(NULL), iOwnsData(EFalse), iSpriteFrame(0)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void RGameSprite::InternalizeL(TResourceReader& aReader)
	{
	Close();
	iId = aReader.ReadInt32();
	if( (!iSpriteTable) || (!iSpriteTable->Exists(iId)) )
		User::Leave(KErrNotFound);
	iSpriteFrame = aReader.ReadInt16();
	iOwnsData = (aReader.ReadUint8() != 0);
	CGameSpriteData* data = new(ELeave) CGameSpriteData;
	CleanupStack::PushL(data);
	data->InternalizeL(aReader);
	if(iOwnsData)
		{
		CleanupStack::Pop(data);
		iData = data;
		}
	else
		{
		CleanupStack::PopAndDestroy(data);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void RGameSprite::InternalizeL(RReadStream& aStream)
	{
	Close();
	iId = aStream.ReadInt32L();
	if( (!iSpriteTable) || (!iSpriteTable->Exists(iId)) )
		User::Leave(KErrNotFound);
	iSpriteFrame = aStream.ReadInt16L();
	iOwnsData = (aStream.ReadUint8L() != 0);
	if(iOwnsData)
		{
		CGameSpriteData* data = new(ELeave) CGameSpriteData;
		CleanupStack::PushL(data);
		data->InternalizeL(aStream);
		CleanupStack::Pop(data);
		iData = data;
		}
	}

///////////////////////////////////////////////////////////////////////////////
void RGameSprite::ExternalizeL(RWriteStream& aStream) const
	{
	aStream.WriteInt32L(iId);
	aStream.WriteInt16L(iSpriteFrame);
	aStream.WriteUint8L(TUint(iOwnsData));
	if(iOwnsData && iData)
		{
		iData->ExternalizeL(aStream);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void RGameSprite::SetSpriteTable(CGameSpriteTable& aSpriteTable)
	{
	Close();
	iSpriteTable = &aSpriteTable;
	}

///////////////////////////////////////////////////////////////////////////////
void RGameSprite::OpenL(TSpriteId aId)
	{
	if( (!iSpriteTable) || (!iSpriteTable->Exists(aId)) )
		User::Leave(KErrNotFound);
	iId = aId;
	}

///////////////////////////////////////////////////////////////////////////////
void RGameSprite::Close()
	{
	iId = -1;
	iSpriteFrame = 0;
	if(iOwnsData)
		{
		delete iData;
		iData = NULL;
		}
	}

///////////////////////////////////////////////////////////////////////////////
TInt RGameSprite::Frame() const
	{
	return iSpriteFrame;
	}

///////////////////////////////////////////////////////////////////////////////
void RGameSprite::SetSpriteFrame(TInt aFrame)
	{
	iSpriteFrame = aFrame;
	const TInt lastFrame = FrameCount() - 1;
	if(iSpriteFrame > lastFrame)
		iSpriteFrame = 0;
	else if(iSpriteFrame < 0)
		iSpriteFrame = lastFrame;
	}

///////////////////////////////////////////////////////////////////////////////
void RGameSprite::SetSpriteData(CGameSpriteData* aSpriteData, TBool aOwned)
	{
	if(iOwnsData)
		{
		delete iData;
		}
	iData = aSpriteData;
	iOwnsData = aOwned;
	}

///////////////////////////////////////////////////////////////////////////////
TSpriteId RGameSprite::Id() const
	{
	return iId;
	}

///////////////////////////////////////////////////////////////////////////////
TUidType RGameSprite::Type() const
	{
	AssertOpen();
	return (*iSpriteTable)[iId].Type();
	}

///////////////////////////////////////////////////////////////////////////////
CGameSpriteTable* RGameSprite::SpriteTable() const
	{
	return iSpriteTable;
	}

///////////////////////////////////////////////////////////////////////////////
CGameSpriteData* RGameSprite::Data() const
	{
	return iData;
	}

///////////////////////////////////////////////////////////////////////////////
TInt RGameSprite::FrameCount() const
	{
	AssertOpen();
	return (*iSpriteTable)[iId].FrameCount(iData);
	}

///////////////////////////////////////////////////////////////////////////////
TSize RGameSprite::FrameSize() const
	{
	AssertOpen();
	return (*iSpriteTable)[iId].FrameSize(iSpriteFrame, iData);
	}

///////////////////////////////////////////////////////////////////////////////
void RGameSprite::DrawFrame(CBitmapContext& aGc, const TPoint& aPos, MMotionBounds* aBounds) const
	{
	AssertOpen();
	(*iSpriteTable)[iId].DrawFrame(iSpriteFrame, aGc, aPos, iData, aBounds);
	}

///////////////////////////////////////////////////////////////////////////////
void RGameSprite::AssertOpen() const
	{
	__ASSERT_ALWAYS( (iSpriteTable) && (iId >= 0), Panic(EPanicSpriteNotOpen));
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// CGameSpriteTable
///////////////////////////////////////////////////////////////////////////////
MGameStreamable* CGameSpriteTable::NewByTypeLC(const TUidType& aType)
	{
	return GameFactory::NewSpriteByTypeLC(aType);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

