///////////////////////////////////////////////////////////////////////////////
//
// Motion.cpp
//
//
//
//
//
///////////////////////////////////////////////////////////////////////////////
#include "Motion.h"
#include "GameResourceUtils.h"
#include <barsread.h>


///////////////////////////////////////////////////////////////////////////////
// TMotionPhysics
///////////////////////////////////////////////////////////////////////////////
TMotionPhysics::TMotionPhysics()
: iMass(1)
	{
	MakeStationary();
	}

///////////////////////////////////////////////////////////////////////////////
TMotionPhysics::TMotionPhysics(const TCoOrds& aPosition, const TVector& aVelocity, const TTrigUnit& aMass)
: iMass(aMass), iPosition(aPosition), iVelocity(aVelocity)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void TMotionPhysics::InternalizeL(TResourceReader& aReader)
	{
	iMass = GameResourceUtils::ReadTTrigUnit(aReader);
	iPosition = GameResourceUtils::ReadTTrigXY(aReader);
	iVelocity = GameResourceUtils::ReadTTrigXY(aReader);
	iAcceleration = GameResourceUtils::ReadTTrigXY(aReader);
	}

///////////////////////////////////////////////////////////////////////////////
void TMotionPhysics::InternalizeL(RReadStream& aStream)
	{
	aStream >> iMass;
	aStream >> iPosition;
	aStream >> iVelocity;
	aStream >> iAcceleration;
	}

///////////////////////////////////////////////////////////////////////////////
void TMotionPhysics::ExternalizeL(RWriteStream& aStream) const
	{
	aStream << iMass;
	aStream << iPosition;
	aStream << iVelocity;
	aStream << iAcceleration;
	}

///////////////////////////////////////////////////////////////////////////////
void TMotionPhysics::ApplyForce(const TVector& aForce)
	{
	iAcceleration += TVector(aForce / iMass);
	}

///////////////////////////////////////////////////////////////////////////////
void TMotionPhysics::MakeStationary()
	{
	const TTrigUnit null;
	iVelocity.Set(null, null);
	iAcceleration.Set(null, null);
	}

///////////////////////////////////////////////////////////////////////////////
void TMotionPhysics::Update(const TSize* aDimensions, MMotionBounds* aBounds)
	{
	iVelocity += iAcceleration;
	iPosition += iVelocity;
	iAcceleration.Zero();
	if( (aBounds) && (aDimensions) )
		{
		aBounds->Adjust(iPosition, *aDimensions, *this);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void TMotionPhysics::SetMass(const TTrigUnit& aMass)
	{
	iMass = aMass;
	}

///////////////////////////////////////////////////////////////////////////////
void TMotionPhysics::SetPosition(const TCoOrds& aPos)
	{
	iPosition = aPos;
	}

///////////////////////////////////////////////////////////////////////////////
void TMotionPhysics::SetVelocity(const TVector& aVelocity)
	{
	iVelocity = aVelocity;
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// MMotionBounds
///////////////////////////////////////////////////////////////////////////////
MMotionBounds::MMotionBounds(const TTrigXY& aExtents, const TTrigUnit& aMaxSpeed)
: iExtents(aExtents), iMaxSpeed(aMaxSpeed)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void MMotionBounds::SetExtents(const TTrigXY& aExtents)
	{
	iExtents = aExtents;
	}

///////////////////////////////////////////////////////////////////////////////
void MMotionBounds::SetMaxSpeed(const TTrigUnit& aMaxSpeed)
	{
	iMaxSpeed = aMaxSpeed;
	}

///////////////////////////////////////////////////////////////////////////////
void MMotionBounds::Adjust(TCoOrds& aPosition, const TTrigXY& aDimensions, TMotionPhysics& aMotion)
	{
	// aPosition is TL.
	const TTrigUnit null(0);
	if(aPosition.iX < null)
		{
		AdjustXOutOfBounds(aPosition.iX, aDimensions.iX, ENegative, aMotion);
		}
	else if(aPosition.iX + aDimensions.iX > iExtents.iX)
		{
		AdjustXOutOfBounds(aPosition.iX, aDimensions.iX, EPositive, aMotion);
		}
	if(aPosition.iY < null)
		{
		AdjustYOutOfBounds(aPosition.iY, aDimensions.iY, ENegative, aMotion);
		}
	else if(aPosition.iY + aDimensions.iY > iExtents.iY)
		{
		AdjustXOutOfBounds(aPosition.iY, aDimensions.iY, EPositive, aMotion);
		}
	AdjustVelocity(aMotion);
	}

///////////////////////////////////////////////////////////////////////////////
void MMotionBounds::AdjustVelocity(TMotionPhysics& aMotion)
	{
	TVector v(aMotion.Velocity());
	TTrigUnit u(v.Magnitude());
	if(u > iMaxSpeed)
		{
		u = iMaxSpeed / u;
		u -= (u / 100);
		v.iX *= u;
		v.iY *= u;
		aMotion.SetVelocity(v);
		}
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// TMotionBoundsWrap
///////////////////////////////////////////////////////////////////////////////
TMotionBoundsWrap::TMotionBoundsWrap(const TTrigXY& aExtents, const TTrigUnit& aMaxSpeed)
: MMotionBounds(aExtents, aMaxSpeed)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void TMotionBoundsWrap::Adjust(TCoOrds& aPosition, const TTrigXY& aDimensions, TMotionPhysics& aMotion)
	{
	// aPosition is TL.
	const TTrigUnit null;
	if(aPosition.iX + aDimensions.iX < null)
		{
		AdjustXOutOfBounds(aPosition.iX, aDimensions.iX, ENegative, aMotion);
		}
	else if(aPosition.iX > iExtents.iX)
		{
		AdjustXOutOfBounds(aPosition.iX, aDimensions.iX, EPositive, aMotion);
		}
	if(aPosition.iY + aDimensions.iY < null)
		{
		AdjustYOutOfBounds(aPosition.iY, aDimensions.iY, ENegative, aMotion);
		}
	else if(aPosition.iY > iExtents.iY)
		{
		AdjustXOutOfBounds(aPosition.iY, aDimensions.iY, EPositive, aMotion);
		}
	AdjustVelocity(aMotion);
	}

///////////////////////////////////////////////////////////////////////////////
void TMotionBoundsWrap::AdjustXOutOfBounds(TTrigUnit& aPos, TTrigUnit aWidth, MMotionBounds::TOutOfBounds aOutOfBounds, TMotionPhysics&)
	{
	if(ENegative == aOutOfBounds)
		aPos = iExtents.iX - aWidth;
	else
		aPos.Zero();
	}

///////////////////////////////////////////////////////////////////////////////
void TMotionBoundsWrap::AdjustYOutOfBounds(TTrigUnit& aPos, TTrigUnit aHeight, MMotionBounds::TOutOfBounds aOutOfBounds, TMotionPhysics&)
	{
	if(ENegative == aOutOfBounds)
		aPos = iExtents.iY - aHeight;
	else
		aPos.Zero();
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// TMotionBoundsClip
///////////////////////////////////////////////////////////////////////////////
TMotionBoundsClip::TMotionBoundsClip(const TTrigXY& aExtents, const TTrigUnit& aMaxSpeed)
: MMotionBounds(aExtents, aMaxSpeed)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void TMotionBoundsClip::AdjustXOutOfBounds(TTrigUnit& aPos, TTrigUnit aWidth, MMotionBounds::TOutOfBounds aOutOfBounds, TMotionPhysics&)
	{
	// Clip to extents.
	if(ENegative == aOutOfBounds)
		aPos.Zero();
	else
		aPos = iExtents.iX - aWidth;
	}

///////////////////////////////////////////////////////////////////////////////
void TMotionBoundsClip::AdjustYOutOfBounds(TTrigUnit& aPos, TTrigUnit aHeight, MMotionBounds::TOutOfBounds aOutOfBounds, TMotionPhysics&)
	{
	// Clip to extents.
	if(ENegative == aOutOfBounds)
		aPos.Zero();
	else
		aPos = iExtents.iY - aHeight;
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// TMotionBoundsBounce
///////////////////////////////////////////////////////////////////////////////
TMotionBoundsBounce::TMotionBoundsBounce(const TTrigXY& aExtents, const TTrigUnit& aMaxSpeed)
: TMotionBoundsClip(aExtents, aMaxSpeed)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void TMotionBoundsBounce::AdjustXOutOfBounds(TTrigUnit& aPos, TTrigUnit aWidth, MMotionBounds::TOutOfBounds aOutOfBounds, TMotionPhysics& aMotion)
	{
	TMotionBoundsClip::AdjustXOutOfBounds(aPos, aWidth, aOutOfBounds, aMotion);
	// Flip velocity X.
	TVector v(aMotion.Velocity());
	v.iX = -v.iX;
	aMotion.SetVelocity(v);
	}

///////////////////////////////////////////////////////////////////////////////
void TMotionBoundsBounce::AdjustYOutOfBounds(TTrigUnit& aPos, TTrigUnit aHeight, MMotionBounds::TOutOfBounds aOutOfBounds, TMotionPhysics& aMotion)
	{
	TMotionBoundsClip::AdjustYOutOfBounds(aPos, aHeight, aOutOfBounds, aMotion);
	// Flip velocity Y.
	TVector v(aMotion.Velocity());
	v.iY = -v.iY;
	aMotion.SetVelocity(v);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

