///////////////////////////////////////////////////////////////////////////////
//
// Motion.h
//
//
//
//
//
///////////////////////////////////////////////////////////////////////////////
#ifndef MOTION_H
#define MOTION_H

#include "Physics.h"

class MMotionBounds;
///////////////////////////////////////////////////////////////////////////////
// TMotionPhysics
///////////////////////////////////////////////////////////////////////////////
// Manages object motion.
// ApplyForce() applies an acceleration to the current motion, determined by
// the object's mass.
// Update() updates the current velocity based on acceleration, and current
// position based on current velocity.
///////////////////////////////////////////////////////////////////////////////
class TMotionPhysics
	{
public:
	TMotionPhysics();
	TMotionPhysics(const TCoOrds& aPosition, const TVector& aVelocity, const TTrigUnit& aMass);

	void InternalizeL(TResourceReader&);
	void InternalizeL(RReadStream&);
	void ExternalizeL(RWriteStream&) const;

	inline TTrigUnit Mass() const;
	inline const TCoOrds& Position() const;
	inline const TVector& Velocity() const;
	void SetMass(const TTrigUnit&);
	void SetPosition(const TCoOrds&);
	void SetVelocity(const TVector&);

	void Update(const TSize* aDimensions=NULL, MMotionBounds* aBounds=NULL);

	void ApplyForce(const TVector&);
	void MakeStationary();

private:
	TTrigUnit iMass;
	TCoOrds iPosition;
	TVector iVelocity;
	TVector iAcceleration;
	};

///////////////////////////////////////////////////////////////////////////////
inline TTrigUnit TMotionPhysics::Mass() const { return iMass; }
inline const TCoOrds& TMotionPhysics::Position() const { return iPosition; }
inline const TVector& TMotionPhysics::Velocity() const { return iVelocity; }

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
// MMotionBounds
///////////////////////////////////////////////////////////////////////////////
// Interface for restrcting an object's motion.
///////////////////////////////////////////////////////////////////////////////
class MMotionBounds
	{
public:
	inline MMotionBounds() {}
	MMotionBounds(const TTrigXY& aExtents, const TTrigUnit& aMaxSpeed);
	void SetExtents(const TTrigXY& aExtents);
	void SetMaxSpeed(const TTrigUnit&);
	inline TTrigXY Extents() const { return iExtents; }
	inline TTrigUnit MaxSpeed() const { return iMaxSpeed; }
	virtual void Adjust(TCoOrds& aPosition, const TTrigXY& aDimensions, TMotionPhysics&);
protected:
	virtual void AdjustVelocity(TMotionPhysics&);
	enum TOutOfBounds { ENegative, EPositive };
	virtual void AdjustXOutOfBounds(TTrigUnit&, TTrigUnit aWidth, TOutOfBounds, TMotionPhysics&) = 0;
	virtual void AdjustYOutOfBounds(TTrigUnit&, TTrigUnit aHeight, TOutOfBounds, TMotionPhysics&) = 0;
protected:
	TTrigXY iExtents;
	TTrigUnit iMaxSpeed;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// TMotionBoundsWrap
///////////////////////////////////////////////////////////////////////////////
// If out of bounds, wrap the position to the other side of the extent.
///////////////////////////////////////////////////////////////////////////////
class TMotionBoundsWrap : public MMotionBounds
	{
public:
	inline TMotionBoundsWrap() {}
	TMotionBoundsWrap(const TTrigXY& aExtents, const TTrigUnit& aMaxSpeed);
	virtual void Adjust(TCoOrds& aPosition, const TTrigXY& aDimensions, TMotionPhysics&);
protected:
	virtual void AdjustXOutOfBounds(TTrigUnit&, TTrigUnit aWidth, MMotionBounds::TOutOfBounds, TMotionPhysics&);
	virtual void AdjustYOutOfBounds(TTrigUnit&, TTrigUnit aHeight, MMotionBounds::TOutOfBounds, TMotionPhysics&);
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// TMotionBoundsClip
///////////////////////////////////////////////////////////////////////////////
// If out of bounds, clip to the extents.
///////////////////////////////////////////////////////////////////////////////
class TMotionBoundsClip : public MMotionBounds
	{
public:
	inline TMotionBoundsClip() {}
	TMotionBoundsClip(const TTrigXY& aExtents, const TTrigUnit& aMaxSpeed);
protected:
	virtual void AdjustXOutOfBounds(TTrigUnit&, TTrigUnit aWidth, MMotionBounds::TOutOfBounds, TMotionPhysics&);
	virtual void AdjustYOutOfBounds(TTrigUnit&, TTrigUnit aHeight, MMotionBounds::TOutOfBounds, TMotionPhysics&);
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// TMotionBoundsBounce
///////////////////////////////////////////////////////////////////////////////
// If out of bounds, flip direction of travel.
///////////////////////////////////////////////////////////////////////////////
class TMotionBoundsBounce : public TMotionBoundsClip
	{
public:
	inline TMotionBoundsBounce() {}
	TMotionBoundsBounce(const TTrigXY& aExtents, const TTrigUnit& aMaxSpeed);
protected:
	virtual void AdjustXOutOfBounds(TTrigUnit&, TTrigUnit aWidth, MMotionBounds::TOutOfBounds, TMotionPhysics&);
	virtual void AdjustYOutOfBounds(TTrigUnit&, TTrigUnit aHeight, MMotionBounds::TOutOfBounds, TMotionPhysics&);
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// MMotionBoid
///////////////////////////////////////////////////////////////////////////////
// This interface class represents an autonomous agent; (a 'boid').
// Each game object is one of these.
///////////////////////////////////////////////////////////////////////////////
class MMotionBoid
	{
public:
	virtual TCoOrds Centre() const = 0;
	virtual TTrigUnit Radius() const = 0;
	virtual TVector Velocity() const = 0;
	};

///////////////////////////////////////////////////////////////////////////////
typedef TArray<const MMotionBoid&> TMotionBoidArray;

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
#endif
